"""Functions to create neuroml cell"""

"""
Copyright (c) 2016-2022, EPFL/Blue Brain Project
 This file is part of BluePyOpt <https://github.com/BlueBrain/BluePyOpt>
 This library is free software; you can redistribute it and/or modify it under
 the terms of the GNU Lesser General Public License version 3.0 as published
 by the Free Software Foundation.
 This library is distributed in the hope that it will be useful, but WITHOUT
 ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
 FOR A PARTICULAR PURPOSE.  See the GNU Lesser General Public License for more
 details.
 You should have received a copy of the GNU Lesser General Public License
 along with this library; if not, write to the Free Software Foundation, Inc.,
 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.
"""

import logging

from neuroml import NeuroMLDocument
from pyneuroml import pynml

from .biophys import get_biophys
from .morphology import add_segment_groups
from .morphology import create_morph_nml

logger = logging.getLogger(__name__)


def create_neuroml_cell(
    bpo_cell,
    release_params,
    skip_channels_copy=False,
    custom_channel_ion=None,
    custom_ion_erevs=None,
):
    """Create the cell.

    Arguments:
        bpo_cell (ephys.CellModel): bluepyopt cell
        release_params (dict): the optimized parameters
        skip_channels_copy (bool): True to skip the copy pasting
            of the neuroml channel files
        custom_channel_ion (dict): dict mapping channel to ion
        custom_ion_erevs (dict): dict mapping ion to erev (reversal potential)

    :returns: name of the cell nml file
    """
    # Create the nml file and add the ion channels
    cell_doc = NeuroMLDocument(id=bpo_cell.name)
    # the network name
    network_filename = f"{bpo_cell.name}.net.nml"

    # Morphology
    logger.info(
        "This will create a cell hoc file in order to create a cell nml file"
    )
    create_morph_nml(bpo_cell, network_filename, release_params)

    # change the network temperature.
    # because the pyneurom.export_to_neuroml2 sets it automatically to 6C.
    network_doc = pynml.read_neuroml2_file(network_filename)
    network = network_doc.networks[0]
    network.temperature = f"{bpo_cell.params['celsius'].value} degC"
    pynml.write_neuroml2_file(
        nml2_doc=network_doc, nml2_file_name=network_filename, validate=True
    )

    # get the cell
    nml_cell_loc = f"{bpo_cell.name}_0_0.cell.nml"
    nml_doc = pynml.read_neuroml2_file(nml_cell_loc)

    cell = nml_doc.cells[0]

    # add segment groups (cell)
    add_segment_groups(cell)

    # get biophys
    bio_prop = get_biophys(
        bpo_cell,
        cell_doc,
        release_params,
        skip_non_uniform=True,
        skip_channels_copy=skip_channels_copy,
        custom_channel_ion=custom_channel_ion,
        custom_ion_erevs=custom_ion_erevs,
    )

    # Append biophys to cell
    cell.biophysical_properties = bio_prop

    # Adding notes
    notes = "This cell was exported from BluePyOpt."
    cell.notes = notes

    # write neuroml cell doc
    cell_doc.cells.append(cell)
    pynml.write_neuroml2_file(
        nml2_doc=cell_doc, nml2_file_name=nml_cell_loc, validate=True
    )

    return nml_cell_loc



"""bluepyopt.stoppingCriteria tests"""


import bluepyopt.stoppingCriteria

import pytest


@pytest.mark.unit
def test_MaxNGen():
    """deapext.stoppingCriteria: Testing MaxNGen"""

    max_gen = 3
    criteria = bluepyopt.deapext.stoppingCriteria.MaxNGen(max_gen)

    assert criteria.criteria_met is False
    criteria.check({"gen": max_gen + 1})
    assert criteria.criteria_met is True
    criteria.reset()
    criteria.check({"gen": max_gen})
    assert criteria.criteria_met is False



"""Create evaluator with LFP related efeatures"""

"""
Copyright (c) 2016-2022, EPFL/Blue Brain Project

 This file is part of BluePyOpt <https://github.com/BlueBrain/BluePyOpt>

 This library is free software; you can redistribute it and/or modify it under
 the terms of the GNU Lesser General Public License version 3.0 as published
 by the Free Software Foundation.

 This library is distributed in the hope that it will be useful, but WITHOUT
 ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
 FOR A PARTICULAR PURPOSE.  See the GNU Lesser General Public License for more
 details.

 You should have received a copy of the GNU Lesser General Public License
 along with this library; if not, write to the Free Software Foundation, Inc.,
 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.
"""

import os
import json
import pathlib

import l5pc_lfpy_model

import bluepyopt.ephys as ephys
from bluepyopt.ephys.extra_features_utils import all_1D_features

config_dir = pathlib.Path(__file__).parents[1] / "l5pc" / "config"


extra_kwargs = dict(
    fs=20,
    fcut=[300, 6000],
    filt_type="filtfilt",
    ms_cut=[3, 5],
    upsample=10
)


def define_protocols():
    """Define protocols"""
    
    protocol_name = "Step1"

    protocol_definition = json.load(open(config_dir / "protocols.json")
                                    )[protocol_name]

    soma_loc = ephys.locations.NrnSeclistCompLocation(
        name="soma", seclist_name="somatic", sec_index=0, comp_x=0.5
    )

    # By default include somatic recording
    somav_recording = ephys.recordings.CompRecording(
        name="%s.soma.v" % protocol_name, location=soma_loc, variable="v"
    )
    mea_recording = ephys.recordings.LFPRecording("%s.MEA.v" % protocol_name)
    
    recordings = [somav_recording, mea_recording]

    stimuli = []
    for stimulus_definition in protocol_definition["stimuli"]:
        stimuli.append(
            ephys.stimuli.LFPySquarePulse(
                step_amplitude=stimulus_definition["amp"],
                step_delay=stimulus_definition["delay"],
                step_duration=stimulus_definition["duration"],
                location=soma_loc,
                total_duration=stimulus_definition["totduration"],
            )
        )

    return {
        protocol_name: ephys.protocols.SweepProtocol(
            protocol_name, stimuli, recordings)
    }


def get_feature_name(protocol_name, location, feature):
    return "%s.%s.%s" % (protocol_name, location, feature)


def get_recording_names(protocol_name, location=None):
    return {"": "%s.%s.v" % (protocol_name, location)}


def define_fitness_calculator(protocols, feature_file):

    with open(feature_file, "r") as f:
        feature_definitions = json.load(f)

    objectives = []
    threshold = -20

    for protocol_name in protocols:
        for location, features in feature_definitions[protocol_name].items():
            recording_names = get_recording_names(protocol_name, location)
            for efel_feature_name, meanstd in features.items():
                feature_name = get_feature_name(protocol_name, location, efel_feature_name)

                stimulus = protocols[protocol_name].stimuli[0]

                args = {
                    "name": feature_name,
                    "stim_start": stimulus.step_delay, 
                    "stim_end": stimulus.step_delay + stimulus.step_duration,
                    "exp_mean": meanstd[0],
                    "exp_std": meanstd[1],
                    "recording_names": recording_names,
                    "threshold": threshold,
                }

                if "MEA" not in location:
                    feature = ephys.efeatures.eFELFeature(
                        efel_feature_name=efel_feature_name,
                        stimulus_current=stimulus.step_amplitude,
                        **args
                    )
                else:
                    somatic_recording_name = recording_names[""].replace("MEA", "soma")
                    feature = ephys.efeatures.extraFELFeature(
                        extrafel_feature_name=efel_feature_name,
                        somatic_recording_name=somatic_recording_name,
                        channel_ids=None,
                        **args,
                        **extra_kwargs
                    )

                objective = ephys.objectives.SingletonObjective(
                    feature_name, feature
                )
                objectives.append(objective)

    return ephys.objectivescalculators.ObjectivesCalculator(objectives)


def create(feature_file="extra_features.json", cvode_active=True, dt=None):
    """Setup"""

    l5pc_cell = l5pc_lfpy_model.create()

    fitness_protocols = define_protocols()
    fitness_calculator = define_fitness_calculator(fitness_protocols, feature_file)

    param_names = [
        param.name for param in l5pc_cell.params.values() if not param.frozen
    ]

    lfpy_sim = ephys.simulators.LFPySimulator(cvode_active=cvode_active, dt=dt)

    return ephys.evaluators.CellEvaluator(
        cell_model=l5pc_cell,
        param_names=param_names,
        fitness_protocols=fitness_protocols,
        fitness_calculator=fitness_calculator,
        sim=lfpy_sim,
    )



"""Single Objective CMA-es class"""

"""
Copyright (c) 2016-2022, EPFL/Blue Brain Project

 This file is part of BluePyOpt <https://github.com/BlueBrain/BluePyOpt>

 This library is free software; you can redistribute it and/or modify it under
 the terms of the GNU Lesser General Public License version 3.0 as published
 by the Free Software Foundation.

 This library is distributed in the hope that it will be useful, but WITHOUT
 ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
 FOR A PARTICULAR PURPOSE.  See the GNU Lesser General Public License for more
 details.

 You should have received a copy of the GNU Lesser General Public License
 along with this library; if not, write to the Free Software Foundation, Inc.,
 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.
"""

# pylint: disable=R0912, R0914

import logging
import numpy
from math import sqrt, log
import copy

from deap import base
from deap import cma

from .stoppingCriteria import (
    MaxNGen,
    Stagnationv2,
    TolHistFun,
    EqualFunVals,
    NoEffectAxis,
    TolUpSigma,
    TolX,
    ConditionCov,
    NoEffectCoor,
)

from . import utils

logger = logging.getLogger("__main__")


class CMA_SO(cma.Strategy):
    """Single objective covariance matrix adaption"""

    def __init__(
        self,
        centroids,
        offspring_size,
        sigma,
        max_ngen,
        IndCreator,
        RandIndCreator,
        map_function=None,
        use_scoop=False,
        use_stagnation_criterion=True,
    ):
        """Constructor

        Args:
            centroid (list): initial guess used as the starting point of
            the CMA-ES
            offspring_size (int): number of offspring individuals in each
                generation
            sigma (float): initial standard deviation of the distribution
            max_ngen (int): total number of generation to run
            IndCreator (fcn): function returning an individual of the pop
            RandIndCreator (fcn): function creating a random individual.
            map_function (map): function used to map (parallelize) the
                evaluation function calls
            use_scoop (bool): use scoop map for parallel computation
            use_stagnation_criterion (bool): whether to use the stagnation
                stopping criterion on top of the maximum generation criterion
        """

        if offspring_size is None:
            lambda_ = int(4 + 3 * log(len(RandIndCreator())))
        else:
            lambda_ = offspring_size

        if centroids is None:
            starter = RandIndCreator()
        else:
            starter = centroids[0]

        cma.Strategy.__init__(self, starter, sigma, lambda_=lambda_)

        self.population = []
        self.problem_size = len(starter)

        self.map_function = map_function
        self.use_scoop = use_scoop

        # Toolbox specific to this CMA-ES
        self.toolbox = base.Toolbox()
        self.toolbox.register("generate", self.generate, IndCreator)
        self.toolbox.register("update", self.update)

        # Set termination conditions
        self.active = True
        if max_ngen <= 0:
            max_ngen = 100 + 50 * (self.problem_size + 3) ** 2 / numpy.sqrt(
                lambda_
            )

        self.stopping_conditions = [
            MaxNGen(max_ngen),
            TolHistFun(lambda_, self.problem_size),
            EqualFunVals(lambda_, self.problem_size),
            NoEffectAxis(self.problem_size),
            TolUpSigma(float(self.sigma)),
            TolX(),
            ConditionCov(),
            NoEffectCoor(),
        ]
        if use_stagnation_criterion:
            self.stopping_conditions.append(
                Stagnationv2(lambda_, self.problem_size)
            )

    def update(self, population):
        """Update the current covariance matrix strategy from the
        population"""

        population.sort(key=lambda ind: ind.fitness.weighted_reduce,
                        reverse=True)

        old_centroid = self.centroid
        self.centroid = numpy.dot(self.weights, population[0:self.mu])

        c_diff = self.centroid - old_centroid

        # Cumulation : update evolution path
        self.ps = (1 - self.cs) * self.ps + sqrt(
            self.cs * (2 - self.cs) * self.mueff
        ) / self.sigma * numpy.dot(
            self.B, (1.0 / self.diagD) * numpy.dot(self.B.T, c_diff)
        )  # noqa

        hsig = float(
            (
                numpy.linalg.norm(self.ps)
                / sqrt(1.0 - (1.0 - self.cs) **
                (2.0 * (self.update_count + 1.0)))
                / self.chiN
                < (1.4 + 2.0 / (self.dim + 1.0))
            )
        )  # noqa

        self.update_count += 1

        self.pc = (1 - self.cc) * self.pc + hsig * sqrt(
            self.cc * (2 - self.cc) * self.mueff
        ) / self.sigma * c_diff

        # Update covariance matrix
        artmp = population[0:self.mu] - old_centroid
        self.C = (
            (
                1
                - self.ccov1
                - self.ccovmu
                + (1 - hsig) * self.ccov1 * self.cc * (2 - self.cc)
            )
            * self.C
            + self.ccov1 * numpy.outer(self.pc, self.pc)
            + self.ccovmu * numpy.dot((self.weights * artmp.T), artmp)
            / self.sigma ** 2
        )

        self.sigma *= numpy.exp(
            (numpy.linalg.norm(self.ps) / self.chiN - 1.0) * self.cs
            / self.damps
        )

        self.diagD, self.B = numpy.linalg.eigh(self.C)
        indx = numpy.argsort(self.diagD)

        self.cond = self.diagD[indx[-1]] / self.diagD[indx[0]]

        self.diagD = self.diagD[indx] ** 0.5
        self.B = self.B[:, indx]
        self.BD = self.B * self.diagD

    def get_population(self, to_space):
        """Returns the population in the original parameter space"""
        pop = copy.deepcopy(self.population)
        for i, ind in enumerate(pop):
            for j, v in enumerate(ind):
                pop[i][j] = to_space[j](v)
        return pop

    def generate_new_pop(self, lbounds, ubounds):
        """Generate a new population bounded in the normalized space"""
        self.population = self.toolbox.generate()
        return utils.bound(self.population, lbounds, ubounds)

    def update_strategy(self):
        self.toolbox.update(self.population)

    def set_fitness(self, fitnesses):
        for f, ind in zip(fitnesses, self.population):
            ind.fitness.values = f

    def check_termination(self, gen):
        stopping_params = {
            "gen": gen,
            "population": self.population,
            "centroid": self.centroid,
            "pc": self.pc,
            "C": self.C,
            "B": self.B,
            "sigma": self.sigma,
            "diagD": self.diagD,
            "cond": self.cond,
        }

        [c.check(stopping_params) for c in self.stopping_conditions]
        for c in self.stopping_conditions:
            if c.criteria_met:
                logger.info(
                    "CMA stopped because of termination criteria: " +
                    "" + " ".join(c.name)
                )
                self.active = False



"""StoppingCriteria class"""

"""
Copyright (c) 2016-2022, EPFL/Blue Brain Project

 This file is part of BluePyOpt <https://github.com/BlueBrain/BluePyOpt>

 This library is free software; you can redistribute it and/or modify it under
 the terms of the GNU Lesser General Public License version 3.0 as published
 by the Free Software Foundation.

 This library is distributed in the hope that it will be useful, but WITHOUT
 ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
 FOR A PARTICULAR PURPOSE.  See the GNU Lesser General Public License for more
 details.

 You should have received a copy of the GNU Lesser General Public License
 along with this library; if not, write to the Free Software Foundation, Inc.,
 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.
"""

# pylint: disable=R0912, R0914

import logging
import numpy

from collections import deque

import bluepyopt.stoppingCriteria

logger = logging.getLogger("__main__")


def isclose(a, b, rel_tol=1e-09, abs_tol=0.0):
    return abs(a - b) <= max(rel_tol * max(abs(a), abs(b)), abs_tol)


class MaxNGen(bluepyopt.stoppingCriteria.StoppingCriteria):
    """Max ngen stopping criteria class"""

    name = "Max ngen"

    def __init__(self, max_ngen):
        """Constructor"""
        super(MaxNGen, self).__init__()
        self.max_ngen = max_ngen

    def check(self, kwargs):
        """Check if the maximum number of iteration is reached"""
        gen = kwargs.get("gen")
        if gen > self.max_ngen:
            self.criteria_met = True


class Stagnation(bluepyopt.stoppingCriteria.StoppingCriteria):
    """Stagnation stopping criteria class"""

    name = "Stagnation"

    def __init__(self, lambda_, problem_size):
        """Constructor"""
        super(Stagnation, self).__init__()

        self.lambda_ = lambda_
        self.problem_size = problem_size
        self.stagnation_iter = None

        self.best = []
        self.median = []

    def check(self, kwargs):
        """Check if the population stopped improving"""
        ngen = kwargs.get("gen")
        population = kwargs.get("population")
        fitness = [ind.fitness.reduce for ind in population]
        fitness.sort()

        # condition to avoid duplicates when re-starting
        if len(self.best) < ngen:
            self.best.append(fitness[0])
            self.median.append(fitness[int(round(len(fitness) / 2.0))])
        self.stagnation_iter = int(
            numpy.ceil(
                0.2 * ngen + 120 + 30.0 * self.problem_size / self.lambda_
            )
        )

        cbest = len(self.best) > self.stagnation_iter
        cmed = len(self.median) > self.stagnation_iter
        cbest2 = numpy.median(self.best[-20:]) >= numpy.median(
            self.best[-self.stagnation_iter:-self.stagnation_iter + 20]
        )
        cmed2 = numpy.median(self.median[-20:]) >= numpy.median(
            self.median[-self.stagnation_iter:-self.stagnation_iter + 20]
        )
        if cbest and cmed and cbest2 and cmed2:
            self.criteria_met = True


class Stagnationv2(bluepyopt.stoppingCriteria.StoppingCriteria):
    """Stagnation stopping criteria class"""

    name = "Stagnationv2"

    def __init__(
        self, lambda_, problem_size, threshold=0.01, std_threshold=0.02
    ):
        """Constructor

        Args:
            lambda_ (int): offspring size
            problem_size (int): problem size
            threshold (float): 1st criterion is triggered if best fitness
                improves less than this threshold for 100 generations
            std_threshold (float): 2nd criterion is triggered if
                standard deviation of the best fitness over
                the last 20 generations is below the best fitness multiplied
                by this threshold
        """
        super(Stagnationv2, self).__init__()

        self.lambda_ = lambda_
        self.problem_size = problem_size
        self.stagnation_iter = None
        self.threshold = threshold
        self.std_threshold = std_threshold

        self.best = []

    def check(self, kwargs):
        """Check if best model fitness does not improve over 1% over 100 gens
            and is not noisy in the last 20 generations
        """
        ngen = kwargs.get("gen")
        population = kwargs.get("population")
        fitness = [ind.fitness.reduce for ind in population]
        fitness.sort()

        # condition to avoid duplicates when re-starting
        if len(self.best) < ngen:
            self.best.append(fitness[0])

        self.stagnation_iter = int(
            numpy.ceil(
                0.2 * ngen + 120 + 30.0 * self.problem_size / self.lambda_
            )
        )

        crit1 = len(self.best) > self.stagnation_iter
        crit2 = numpy.median(self.best[-20:]) * (1 + self.threshold) \
            > numpy.median(self.best[-120:-100])
        crit3 = numpy.std(self.best[-20:]) < (
            self.std_threshold * self.best[-1]
        )

        if crit1 and crit2 and crit3:
            self.criteria_met = True


class TolHistFun(bluepyopt.stoppingCriteria.StoppingCriteria):
    """TolHistFun stopping criteria class"""

    name = "TolHistFun"

    def __init__(self, lambda_, problem_size):
        """Constructor"""
        super(TolHistFun, self).__init__()
        self.tolhistfun = 10 ** -12
        self.mins = deque(
            maxlen=10 + int(numpy.ceil(30.0 * problem_size / lambda_)))

    def check(self, kwargs):
        """Check if the range of the best values is smaller than
        the threshold"""
        population = kwargs.get("population")
        self.mins.append(numpy.min([ind.fitness.reduce for ind in population]))

        if (
            len(self.mins) == self.mins.maxlen
            and max(self.mins) - min(self.mins) < self.tolhistfun
        ):
            self.criteria_met = True


class EqualFunVals(bluepyopt.stoppingCriteria.StoppingCriteria):
    """EqualFunVals stopping criteria class"""

    name = "EqualFunVals"

    def __init__(self, lambda_, problem_size):
        """Constructor"""
        super(EqualFunVals, self).__init__()
        self.problem_size = problem_size
        self.equalvals = float(problem_size) / 3.0
        self.equalvals_k = int(numpy.ceil(0.1 + lambda_ / 4.0))
        self.equalvalues = []

    def check(self, kwargs):
        """Check if in 1/3rd of the last problem_size iterations the best and
        k'th best solutions are equal"""
        ngen = kwargs.get("gen")
        population = kwargs.get("population")

        fitness = [ind.fitness.reduce for ind in population]
        fitness.sort()

        if isclose(fitness[0], fitness[-self.equalvals_k], rel_tol=1e-6):
            self.equalvalues.append(1)
        else:
            self.equalvalues.append(0)

        if (
            ngen > self.problem_size
            and sum(self.equalvalues[-self.problem_size:]) > self.equalvals
        ):
            self.criteria_met = True


class TolX(bluepyopt.stoppingCriteria.StoppingCriteria):
    """TolX stopping criteria class"""

    name = "TolX"

    def __init__(self):
        """Constructor"""
        super(TolX, self).__init__()
        self.tolx = 10 ** -12

    def check(self, kwargs):
        """Check if all components of pc and sqrt(diag(C)) are smaller than
        a threshold"""
        pc = kwargs.get("pc")
        C = kwargs.get("C")

        if all(pc < self.tolx) and all(numpy.sqrt(numpy.diag(C)) < self.tolx):
            self.criteria_met = True


class TolUpSigma(bluepyopt.stoppingCriteria.StoppingCriteria):
    """TolUpSigma stopping criteria class"""

    name = "TolUpSigma"

    def __init__(self, sigma0):
        """Constructor"""
        super(TolUpSigma, self).__init__()
        self.sigma0 = sigma0
        self.tolupsigma = 10 ** 20

    def check(self, kwargs):
        """Check if the sigma/sigma0 ratio is bigger than a threshold"""
        sigma = kwargs.get("sigma")
        diagD = kwargs.get("diagD")

        if sigma / self.sigma0 > float(diagD[-1] ** 2) * self.tolupsigma:
            self.criteria_met = True


class ConditionCov(bluepyopt.stoppingCriteria.StoppingCriteria):
    """ConditionCov stopping criteria class"""

    name = "ConditionCov"

    def __init__(self):
        """Constructor"""
        super(ConditionCov, self).__init__()
        self.conditioncov = 10 ** 14

    def check(self, kwargs):
        """Check if the condition number of the covariance matrix is
        too large"""
        cond = kwargs.get("cond")

        if cond > self.conditioncov:
            self.criteria_met = True


class NoEffectAxis(bluepyopt.stoppingCriteria.StoppingCriteria):
    """NoEffectAxis stopping criteria class"""

    name = "NoEffectAxis"

    def __init__(self, problem_size):
        """Constructor"""
        super(NoEffectAxis, self).__init__()
        self.conditioncov = 10 ** 14
        self.problem_size = problem_size

    def check(self, kwargs):
        """Check if the coordinate axis std is too low"""
        ngen = kwargs.get("gen")
        centroid = kwargs.get("centroid")
        sigma = kwargs.get("sigma")
        diagD = kwargs.get("diagD")
        B = kwargs.get("B")

        noeffectaxis_index = ngen % self.problem_size

        if all(
            centroid
            == centroid
            + 0.1 * sigma * diagD[-noeffectaxis_index] * B[-noeffectaxis_index]
        ):
            self.criteria_met = True


class NoEffectCoor(bluepyopt.stoppingCriteria.StoppingCriteria):
    """NoEffectCoor stopping criteria class"""

    name = "NoEffectCoor"

    def __init__(self):
        """Constructor"""
        super(NoEffectCoor, self).__init__()

    def check(self, kwargs):
        """Check if main axis std has no effect"""
        centroid = kwargs.get("centroid")
        sigma = kwargs.get("sigma")
        C = kwargs.get("C")

        if any(centroid == centroid + 0.2 * sigma * numpy.diag(C)):
            self.criteria_met = True



"""Unit tests for acc."""

from bluepyopt.ephys.acc import arbor, ArbLabel


import pytest


@pytest.mark.unit
def test_arbor_labels():
    """Test Arbor labels."""

    region_label = ArbLabel(type='region',
                            name='first_branch',
                            s_expr='(branch 0)')

    assert region_label.defn == '(region-def "first_branch" (branch 0))'
    assert region_label.ref == '(region "first_branch")'
    assert region_label.name == 'first_branch'
    assert region_label.loc == '(branch 0)'
    assert region_label == region_label
    assert region_label is not None

    locset_label = ArbLabel(type='locset',
                            name='first_branch_center',
                            s_expr='(location 0 0.5)')

    assert locset_label.defn == \
        '(locset-def "first_branch_center" (location 0 0.5))'
    assert locset_label.ref == '(locset "first_branch_center")'
    assert locset_label.name == 'first_branch_center'
    assert locset_label.loc == '(location 0 0.5)'
    assert locset_label == locset_label
    assert locset_label is not None

    assert locset_label != region_label

    arbor.label_dict({region_label.name: region_label.loc,
                      locset_label.name: locset_label.loc})



"""bluepyopt.utils tests"""

import multiprocessing
import time

import bluepyopt.deapext.utils as utils
import pytest


def flag(event):
    """Send a multiprocessing event."""
    time.sleep(1)
    event.set()


def catch_event(event):
    """Verify that run_next_gen changes when event is caught."""
    # None case
    assert utils.run_next_gen(True, None)

    # event is not set case
    assert utils.run_next_gen(True, event)

    # event is set by another process case
    time.sleep(2)
    assert not (utils.run_next_gen(True, event))


@pytest.mark.unit
def test_run_next_gen_condition():
    """deapext.utils: Testing run_next_gen."""
    event = multiprocessing.Event()
    p1 = multiprocessing.Process(target=catch_event, args=(event,))
    p2 = multiprocessing.Process(target=flag, args=(event,))

    p1.start()
    p2.start()

    p1.join()
    p2.join()



"""EPhys test utils"""

from bluepyopt import ephys
from bluepyopt.ephys.parameters import (
    NrnGlobalParameter, NrnSectionParameter, NrnRangeParameter,)
from bluepyopt.ephys.locations import NrnSeclistLocation


def make_mech():
    """Create mechanism"""
    basal = ephys.locations.NrnSeclistLocation('basal', seclist_name='basal')
    apical = ephys.locations.NrnSeclistLocation(
        'apical', seclist_name='apical')
    return ephys.mechanisms.NrnMODMechanism(
        'Ih',
        suffix='Ih',
        locations=[
            basal,
            apical,
        ])


def make_parameters():
    """Create parameters"""
    value, frozen, bounds, param_name = 65, False, [
        0, 100.0], 'gSKv3_1bar_SKv3_1'
    value_scaler = ephys.parameterscalers.NrnSegmentLinearScaler()
    locations = (NrnSeclistLocation('Location0', 'somatic'),
                 NrnSeclistLocation('Location1', 'apical'),)
    parameters = (
        NrnGlobalParameter(
            'NrnGlobalParameter',
            value,
            frozen,
            bounds,
            param_name),
        NrnSectionParameter(
            'NrnSectionParameter',
            value,
            frozen,
            bounds,
            param_name,
            value_scaler,
            locations),
        NrnRangeParameter(
            'NrnRangeParameter',
            value,
            frozen,
            bounds,
            param_name,
            value_scaler,
            locations),
    )
    return parameters



"""Simulator classes"""

# pylint: disable=W0511

import os
import logging
import imp
import ctypes
import platform
import warnings
import pathlib

from bluepyopt.ephys.acc import arbor


logger = logging.getLogger(__name__)


class NrnSimulator(object):

    """Neuron simulator"""

    def __init__(self, dt=None, cvode_active=True, cvode_minstep=None,
                 random123_globalindex=None, mechanisms_directory=None):
        """Constructor

        Args:
            dt (float): the integration time step used by Neuron.
            cvode_active (bool): should Neuron use the variable time step
                integration method
            cvode_minstep (float): the minimum time step allowed for a cvode
                step. Default is 0.0.
            random123_globalindex (int): used to set the global index used by
                all instances of the Random123 instances of Random
            mechanisms_directory (str): path to the parent directory of the
                directory containing the mod files. If the mod files are in
                "./data/mechanisms", then mechanisms_directory should be
                "./data/".
        """

        # hoc.so does not exist on NEURON Windows or MacOS
        # although \\hoc.pyd can work here, it gives an error for
        # nrn_nobanner_ line
        self.disable_banner = platform.system() not in ['Windows', 'Darwin']
        self.banner_disabled = False
        self.mechanisms_directory = mechanisms_directory

        self.dt = dt if dt is not None else self.neuron.h.dt

        self.cvode_minstep_value = cvode_minstep

        self.cvode_active = cvode_active

        self.initialize()

        self.random123_globalindex = random123_globalindex

    @property
    def cvode(self):
        """Return cvode instance"""

        return self.neuron.h.CVode()

    @property
    def cvode_minstep(self):
        """Return cvode minstep value"""

        return self.cvode.minstep()

    @cvode_minstep.setter
    def cvode_minstep(self, value):
        """Set cvode minstep value"""

        self.cvode.minstep(value)

    @staticmethod
    def _nrn_disable_banner():
        """Disable Neuron banner"""

        nrnpy_path = os.path.join(imp.find_module('neuron')[1])
        import glob
        hoc_so_list = \
            glob.glob(os.path.join(nrnpy_path, 'hoc*.so'))

        if len(hoc_so_list) != 1:
            warnings.warn('Unable to find Neuron hoc shared library in %s, '
                          'not disabling banner' % nrnpy_path)
        else:
            hoc_so = hoc_so_list[0]
            nrndll = ctypes.cdll[hoc_so]
            ctypes.c_int.in_dll(nrndll, 'nrn_nobanner_').value = 1

    # pylint: disable=R0201
    @property
    def neuron(self):
        """Return Neuron module"""

        if self.disable_banner and not self.banner_disabled:
            NrnSimulator._nrn_disable_banner()
            self.banner_disabled = True

        import neuron  # NOQA

        if self.mechanisms_directory is not None:
            neuron.load_mechanisms(
                self.mechanisms_directory, warn_if_already_loaded=False
            )

        return neuron

    def initialize(self):
        """Initialize simulator: Set Neuron variables"""
        self.neuron.h.load_file('stdrun.hoc')
        self.neuron.h.dt = self.dt
        self.neuron.h.cvode_active(1 if self.cvode_active else 0)

    def run(
            self,
            tstop=None,
            dt=None,
            cvode_active=None,
            random123_globalindex=None):
        """Run protocol"""

        self.neuron.h.tstop = tstop

        if cvode_active and dt is not None:
            raise ValueError(
                'NrnSimulator: Impossible to combine the dt argument when '
                'cvode_active is True in the NrnSimulator run method')

        if cvode_active is None:
            cvode_active = self.cvode_active

        if not cvode_active and dt is None:  # use dt of simulator
            if self.neuron.h.dt != self.dt:
                raise Exception(
                    'NrnSimulator: Some process has changed the '
                    'time step dt of Neuron since the creation of this '
                    'NrnSimulator object. Not sure this is intended:\n'
                    'current dt: %.6g\n'
                    'init dt: %.6g' % (self.neuron.h.dt, self.dt))
            dt = self.dt

        self.neuron.h.cvode_active(1 if cvode_active else 0)
        if self.cvode_minstep_value is not None:
            save_minstep = self.cvode_minstep
            self.cvode_minstep = self.cvode_minstep_value

        if cvode_active:
            logger.debug('Running Neuron simulator %.6g ms, with cvode', tstop)
        else:
            self.neuron.h.dt = dt
            self.neuron.h.steps_per_ms = 1.0 / dt
            logger.debug(
                'Running Neuron simulator %.6g ms, with dt=%r',
                tstop,
                dt)

        if random123_globalindex is None:
            random123_globalindex = self.random123_globalindex

        if random123_globalindex is not None:
            rng = self.neuron.h.Random()
            rng.Random123_globalindex(random123_globalindex)

        try:
            self.neuron.h.run()
        except Exception as e:
            raise NrnSimulatorException('Neuron simulator error', e)

        if self.cvode_minstep_value is not None:
            self.cvode_minstep = save_minstep

        logger.debug('Neuron simulation finished')


class NrnSimulatorException(Exception):

    """All exception generated by Neuron simulator"""

    def __init__(self, message, original):
        """Constructor"""

        super(NrnSimulatorException, self).__init__(message)
        self.original = original


class LFPySimulator(NrnSimulator):

    """LFPy simulator"""

    def __init__(self, dt=None, cvode_active=True, cvode_minstep=None,
                 random123_globalindex=None, mechanisms_directory=None):
        """Constructor

        Args:
            dt (float): the integration time step used by neuron.
            cvode_active (bool): should neuron use the variable time step
                integration method
            cvode_minstep (float): the minimum time step allowed for a cvode
                step. Default is 0.0.
            random123_globalindex (int): used to set the global index used by
                all instances of the Random123 instances of Random
            mechanisms_directory (str): path to the parent directory of the
                directory containing the mod files. If the mod files are in
                "./data/mechanisms", then mechanisms_directory should be
                "./data/".
        """

        super(LFPySimulator, self).__init__(
            dt=dt,
            cvode_active=cvode_active,
            cvode_minstep=cvode_minstep,
            random123_globalindex=random123_globalindex,
            mechanisms_directory=mechanisms_directory
        )

    def run(
            self,
            lfpy_cell,
            lfpy_electrode,
            tstop=None,
            dt=None,
            cvode_active=None,
            random123_globalindex=None):
        """Run protocol"""
        _ = self.neuron

        lfpy_cell.tstart = 0.0
        lfpy_cell.tstop = tstop

        if dt is not None:
            lfpy_cell.dt = dt

        if cvode_active and dt is not None:
            raise ValueError(
                'NrnSimulator: Impossible to combine the dt argument when '
                'cvode_active is True in the NrnSimulator run method')

        if cvode_active is None:
            cvode_active = self.cvode_active

        if cvode_active is not None:
            self.cvode_active = cvode_active

        if random123_globalindex is None:
            random123_globalindex = self.random123_globalindex

        if random123_globalindex is not None:
            rng = self.neuron.h.Random()
            rng.Random123_globalindex(random123_globalindex)

        probes = [lfpy_electrode] if lfpy_electrode is not None else None

        sim_params = {
            "probes": probes,
            "rec_vmem": False,
            "rec_imem": False,
            "rec_ipas": False,
            "rec_icap": False,
            "rec_variables": [],
            "variable_dt": self.cvode_active,
            "atol": 0.001,
            "to_memory": True,
            "to_file": False,
            "file_name": None,
        }

        try:
            lfpy_cell.simulate(**sim_params)
        except Exception as e:
            raise LFPySimulatorException("LFPy simulator error", e)

        logger.debug("LFPy simulation finished")


class LFPySimulatorException(Exception):

    """All exception generated by LFPy simulator"""

    def __init__(self, message, original):
        """Constructor"""

        super(LFPySimulatorException, self).__init__(message)
        self.original = original


class ArbSimulator(object):

    """Arbor simulator"""

    def __init__(self, dt=None, ext_catalogues=None):
        """Constructor

        Args:
            dt (float): the integration time step used by Arbor.
            ext_catalogues (): Name to path mapping of non-Arbor built-in
            NMODL mechanism catalogues compiled with modcc
        """

        self.dt = dt
        self.ext_catalogues = ext_catalogues
        if ext_catalogues is not None:
            for cat, cat_path in ext_catalogues.items():
                cat_lib = '%s-catalogue.so' % cat
                cat_path = pathlib.Path(cat_path).resolve()
                if not os.path.exists(cat_path / cat_lib):
                    raise ArbSimulatorException(
                        'Cannot find %s at %s - first build' % (cat_lib,
                                                                cat_path) +
                        ' mechanism catalogue with modcc:' +
                        ' arbor-build-catalogue %s %s' % (cat, cat_path))
        # TODO: add parameters for discretization

    def initialize(self):
        """Initialize simulator"""
        pass

    def instantiate(self, morph, decor, labels):
        cable_cell = arbor.cable_cell(morphology=morph,
                                      decor=decor,
                                      labels=labels)

        arb_cell_model = arbor.single_cell_model(cable_cell)

        # Add catalogues with explicit qualifiers
        arb_cell_model.properties.catalogue = arbor.catalogue()

        # User-supplied catalogues take precedence
        if self.ext_catalogues is not None:
            for cat, cat_path in self.ext_catalogues.items():
                cat_lib = '%s-catalogue.so' % cat
                cat_path = pathlib.Path(cat_path).resolve()
                arb_cell_model.properties.catalogue.extend(
                    arbor.load_catalogue(cat_path / cat_lib),
                    cat + "::")

        # Built-in catalogues are always added (could be made optional)
        if self.ext_catalogues is None or \
                'default' not in self.ext_catalogues:
            arb_cell_model.properties.catalogue.extend(
                arbor.default_catalogue(), "default::")

        if self.ext_catalogues is None or \
                'BBP' not in self.ext_catalogues:
            arb_cell_model.properties.catalogue.extend(
                arbor.bbp_catalogue(), "BBP::")

        if self.ext_catalogues is None or \
                'allen' not in self.ext_catalogues:
            arb_cell_model.properties.catalogue.extend(
                arbor.allen_catalogue(), "allen::")

        return arb_cell_model

    def run(self,
            arb_cell_model,
            tstop=None,
            dt=None):

        dt = dt if dt is not None else self.dt

        if dt is not None:
            return arb_cell_model.run(tfinal=tstop, dt=dt)
        else:
            return arb_cell_model.run(tfinal=tstop)


class ArbSimulatorException(Exception):

    """All exception generated by Arbor simulator"""

    def __init__(self, message):
        """Constructor"""

        super(ArbSimulatorException, self).__init__(message)



"""bluepyopt.ephys.simulators tests"""

"""
Copyright (c) 2016-2020, EPFL/Blue Brain Project

 This file is part of BluePyOpt <https://github.com/BlueBrain/BluePyOpt>

 This library is free software; you can redistribute it and/or modify it under
 the terms of the GNU Lesser General Public License version 3.0 as published
 by the Free Software Foundation.

 This library is distributed in the hope that it will be useful, but WITHOUT
 ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
 FOR A PARTICULAR PURPOSE.  See the GNU Lesser General Public License for more
 details.

 You should have received a copy of the GNU Lesser General Public License
 along with this library; if not, write to the Free Software Foundation, Inc.,
 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.
"""

# pylint:disable=W0612

import numpy


import pytest
import numpy

import bluepyopt.ephys as ephys
from .testmodels import dummycells


@pytest.mark.unit
def test_stimulus_init():
    """ephys.stimuli: test if Stimulus constructor works"""

    stim = ephys.stimuli.Stimulus()
    assert isinstance(stim, ephys.stimuli.Stimulus)


@pytest.mark.unit
def test_NrnNetStimStimulus_init():
    """ephys.stimuli: test if NrnNetStimStimulus constructor works"""

    pytest.raises(ValueError, ephys.stimuli.NrnNetStimStimulus)

    stim = ephys.stimuli.NrnNetStimStimulus(total_duration=100)
    assert isinstance(stim, ephys.stimuli.NrnNetStimStimulus)

    assert str(stim) == 'Netstim'


@pytest.mark.unit
def test_NrnNetStimStimulus_instantiate():
    """ephys.stimuli: test if NrnNetStimStimulus instantiate works"""

    nrn_sim = ephys.simulators.NrnSimulator()
    dummy_cell = dummycells.DummyCellModel1()
    icell = dummy_cell.instantiate(sim=nrn_sim)

    somacenter_loc = ephys.locations.NrnSeclistCompLocation(
        name=None,
        seclist_name='somatic',
        sec_index=0,
        comp_x=.5)

    expsyn_mech = ephys.mechanisms.NrnMODPointProcessMechanism(
        name='expsyn',
        suffix='ExpSyn',
        locations=[somacenter_loc])

    expsyn_mech.instantiate(sim=nrn_sim, icell=icell)

    expsyn_loc = ephys.locations.NrnPointProcessLocation(
        'expsyn_loc',
        pprocess_mech=expsyn_mech)

    netstim = ephys.stimuli.NrnNetStimStimulus(
        total_duration=200,
        number=5,
        interval=5,
        start=20,
        weight=5e-4,
        locations=[expsyn_loc])

    netstim.instantiate(sim=nrn_sim, icell=icell)

    nrn_sim.run(netstim.total_duration)

    expsyn_mech.destroy(sim=nrn_sim)
    netstim.destroy(sim=nrn_sim)
    dummy_cell.destroy(sim=nrn_sim)


@pytest.mark.unit
def test_NrnCurrentPlayStimulus_instantiate():
    """ephys.stimuli: test if NrnNetStimStimulus instantiate works"""

    nrn_sim = ephys.simulators.NrnSimulator()
    dummy_cell = dummycells.DummyCellModel1()
    icell = dummy_cell.instantiate(sim=nrn_sim)

    somacenter_loc = ephys.locations.NrnSeclistCompLocation(
        name=None,
        seclist_name='somatic',
        sec_index=0,
        comp_x=.5)

    time_points = [10, 50]
    current_points = [0.1, 0.2]
    current_stim = ephys.stimuli.NrnCurrentPlayStimulus(
        time_points=time_points,
        current_points=current_points,
        location=somacenter_loc)

    assert current_stim.time_points == time_points
    assert current_stim.current_points == current_points
    assert str(current_stim) == 'Current play at somatic[0](0.5)'
    current_stim.instantiate(sim=nrn_sim, icell=icell)

    nrn_sim.run(100)

    current_stim.destroy(sim=nrn_sim)
    dummy_cell.destroy(sim=nrn_sim)


@pytest.mark.unit
def test_NrnRampPulse_init():
    """ephys.stimuli: test if NrnRampPulse constructor works"""
    stim = ephys.stimuli.NrnRampPulse()
    assert isinstance(stim, ephys.stimuli.NrnRampPulse)


@pytest.mark.unit
def test_NrnRampPulse_instantiate():
    """ephys.stimuli: test if NrnRampPulse injects correct current"""

    nrn_sim = ephys.simulators.NrnSimulator()
    dummy_cell = dummycells.DummyCellModel1()
    icell = dummy_cell.instantiate(sim=nrn_sim)
    soma_loc = ephys.locations.NrnSeclistCompLocation(
        name=None,
        seclist_name='somatic',
        sec_index=0,
        comp_x=.5)
    recording = ephys.recordings.CompRecording(
        location=soma_loc,
        variable='v')

    ramp_amplitude_start = 0.1
    ramp_amplitude_end = 1.0
    ramp_delay = 20.0
    ramp_duration = 20.0
    total_duration = 50.0

    stim = ephys.stimuli.NrnRampPulse(
        ramp_amplitude_start=ramp_amplitude_start,
        ramp_amplitude_end=ramp_amplitude_end,
        ramp_delay=ramp_delay,
        ramp_duration=ramp_duration,
        total_duration=total_duration,
        location=soma_loc)

    assert (
        str(stim) ==
        'Ramp pulse amp_start 0.100000 amp_end 1.000000 '
        'delay 20.000000 duration 20.000000 totdur 50.000000'
        ' at somatic[0](0.5)')
    stim.instantiate(sim=nrn_sim, icell=icell)

    recording.instantiate(sim=nrn_sim, icell=icell)

    stim_i_vec = nrn_sim.neuron.h.Vector()
    stim_i_vec.record(stim.iclamp._ref_i)  # pylint: disable=W0212
    nrn_sim.run(stim.total_duration)

    current = numpy.array(stim_i_vec.to_python())
    time = numpy.array(recording.response['time'])
    voltage = numpy.array(recording.response['voltage'])

    # make sure current is 0 before stimulus
    assert numpy.max(
        current[numpy.where((0 <= time) & (time < ramp_delay))]) == 0

    # make sure voltage stays at v_init before stimulus
    assert numpy.max(
        voltage[
            numpy.where((0 <= time)
                        & (time < ramp_delay))]) == nrn_sim.neuron.h.v_init

    # make sure current is at right amp at end of stimulus
    assert (
        current[numpy.where(time == ramp_delay)][-1] ==
        ramp_amplitude_start)
    # make sure current is at right amp at end of stimulus
    assert (
        current[numpy.where(time == (ramp_delay + ramp_duration))][0] ==
        ramp_amplitude_end)

    # make sure current is 0 after stimulus
    assert numpy.max(
        current[
            numpy.where(
                (ramp_delay + ramp_duration < time)
                & (time <= total_duration))]) == 0

    # make sure voltage is correct after stimulus
    numpy.testing.assert_almost_equal(numpy.mean(
        voltage[
            numpy.where(
                (ramp_delay + ramp_duration < time)
                & (time <= total_duration))]), -57.994437612124869)
    recording.destroy(sim=nrn_sim)
    stim.destroy(sim=nrn_sim)
    dummy_cell.destroy(sim=nrn_sim)


@pytest.mark.unit
def test_LFPySquarePulse_init():
    """ephys.stimuli: test if LFPySquarePulse constructor works"""
    soma_loc = ephys.locations.NrnSeclistCompLocation(
        name=None,
        seclist_name='somatic',
        sec_index=0,
        comp_x=.5)
    stim = ephys.stimuli.LFPySquarePulse(
        step_amplitude=0.1,
        step_delay=70,
        step_duration=100,
        location=soma_loc,
        total_duration=300
    )
    assert isinstance(stim, ephys.stimuli.LFPySquarePulse)
    assert stim.step_amplitude == 0.1
    assert stim.step_delay == 70
    assert stim.step_duration == 100
    assert isinstance(stim.location, ephys.locations.NrnSeclistCompLocation)
    assert stim.total_duration == 300

    assert str(stim) == (
        "Square pulse amp 0.100000 delay 70.000000 duration 100.000000 "
        "totdur 300.000000 at somatic[0](0.5)"
    )


@pytest.mark.unit
def test_LFPySquarePulse_instantiate():
    """ephys.stimuli: test if LFPySquarePulse instantiate works"""

    lfpy_sim = ephys.simulators.LFPySimulator()
    dummy_cell = dummycells.DummyLFPyCellModel1()
    _, lfpy_cell = dummy_cell.instantiate(sim=lfpy_sim)

    soma_loc = ephys.locations.NrnSeclistCompLocation(
        name=None,
        seclist_name='somatic',
        sec_index=0,
        comp_x=.5)

    stim = ephys.stimuli.LFPySquarePulse(
        step_amplitude=0.1,
        step_delay=70,
        step_duration=100,
        location=soma_loc,
        total_duration=300
    )

    stim.instantiate(sim=lfpy_sim, lfpy_cell=lfpy_cell)
    lfpy_sim.run(
        lfpy_cell=lfpy_cell,
        lfpy_electrode=dummy_cell.lfpy_electrode,
        tstop=stim.total_duration
    )

    stim.destroy(sim=lfpy_sim)
    dummy_cell.destroy(sim=lfpy_sim)



'''create a hoc file from a set of BluePyOpt.ephys parameters'''

# pylint: disable=R0914

import os
import re

from collections import defaultdict, namedtuple, OrderedDict
from datetime import datetime

import jinja2
import bluepyopt
from bluepyopt.ephys.locations import (NrnSeclistCompLocation,
                                       NrnSeclistLocation,
                                       NrnSectionCompLocation,
                                       NrnSomaDistanceCompLocation,
                                       NrnSecSomaDistanceCompLocation,
                                       NrnTrunkSomaDistanceCompLocation,
                                       ArbLocation)

from bluepyopt.ephys.mechanisms import (Mechanism,
                                        NrnMODMechanism,
                                        NrnMODPointProcessMechanism)

from bluepyopt.ephys.parameters import (NrnGlobalParameter,
                                        NrnSectionParameter,
                                        NrnRangeParameter,
                                        NrnPointProcessParameter,
                                        MetaParameter)

from bluepyopt.ephys.parameterscalers import (NrnSegmentSomaDistanceScaler,
                                              NrnSegmentLinearScaler,
                                              FLOAT_FORMAT,
                                              format_float)

PointExpr = namedtuple('PointExpr', 'name, point_loc, value')
RangeExpr = namedtuple('RangeExpr', 'location, name, value, value_scaler')

# Consider renaming Location as name already used in locations module
Location = namedtuple('Location', 'name, value')
Range = namedtuple('Range', 'location, param_name, value')

DEFAULT_LOCATION_ORDER = [
    'all',
    'apical',
    'axonal',
    'basal',
    'somatic',
    'myelinated']


def generate_channels_by_location(mechs, location_order):
    """Create a OrderedDictionary of all channel mechs for hoc template.

    Args:
        mechs (list of bluepyopt.ephys.mechanisms.Mechanism): mechanisms
        location_order (list of str): order of locations

    Returns: tuple of channels, point_channels and location order
    """
    loc_desc = _loc_desc
    return _generate_channels_by_location(mechs, location_order, loc_desc)


def _generate_channels_by_location(mechs, location_order, loc_desc):
    """Create a OrderedDictionary of all channel mechs for hoc template."""
    channels = OrderedDict((location, []) for location in location_order)
    point_channels = OrderedDict((location, []) for location in location_order)
    for mech in mechs:
        name = mech.suffix
        for location in mech.locations:
            if isinstance(mech, NrnMODPointProcessMechanism):
                point_channels[loc_desc(location, mech)].append(mech)
            else:
                channels[loc_desc(location, mech)].append(name)
    return channels, point_channels


def generate_reinitrng(mechs) -> str:
    """Create re_init_rng function"""

    for mech in mechs:
        if isinstance(mech, NrnMODPointProcessMechanism):
            raise NotImplementedError(
                'HOC generation for models with point process mechanisms'
                ' is not yet supported.')

    reinitrng_hoc_blocks = ''

    for mech in mechs:
        reinitrng_hoc_blocks += mech.generate_reinitrng_hoc_block()

    reinitrng_content = NrnMODMechanism.hash_hoc_string

    reinitrng_content += NrnMODMechanism.reinitrng_hoc_string % {
        'reinitrng_hoc_blocks': reinitrng_hoc_blocks}

    return reinitrng_content


def range_exprs_to_hoc(range_params):
    """Process raw range parameters to hoc strings"""

    ret = []
    for param in range_params:
        value = param.value_scaler.inst_distribution
        value = re.sub(r'math\.', '', value)
        value = re.sub('{distance}', FLOAT_FORMAT, value)
        value = re.sub('{value}', format_float(param.value), value)
        ret.append(Range(param.location, param.name, value))
    return ret


def _loc_desc(location, param_or_mech):
    """Generate Neuron location description for HOC template"""

    if isinstance(param_or_mech, Mechanism):
        if isinstance(param_or_mech, NrnMODMechanism):
            if isinstance(location, NrnSeclistLocation):
                return location.seclist_name
            else:
                raise CreateHocException(
                    "%s is currently not supported for mechs." %
                    type(location).__name__)
        elif isinstance(param_or_mech, NrnMODPointProcessMechanism):
            raise CreateHocException("%s is currently not supported." %
                                     type(param_or_mech).__name__)
    elif not isinstance(location, (NrnSeclistCompLocation,
                                   NrnSectionCompLocation,
                                   NrnSomaDistanceCompLocation,
                                   NrnSecSomaDistanceCompLocation,
                                   NrnTrunkSomaDistanceCompLocation,
                                   ArbLocation)) and \
            not isinstance(param_or_mech, NrnPointProcessParameter):
        return location.seclist_name
    else:
        raise CreateHocException("%s is currently not supported." %
                                 type(param_or_mech).__name__)


def generate_parameters(parameters):
    """Create a list of parameters that need to be added to the hoc template

    Args:
        parameters (list of bluepyopt.Parameters): parameters in hoc template

    Returns: tuple of global, section, range, pprocess and location order
    """
    location_order = DEFAULT_LOCATION_ORDER
    loc_desc = _loc_desc
    return _generate_parameters(parameters, location_order, loc_desc)


def _generate_parameters(parameters, location_order, loc_desc):
    """Create a list of parameters that need to be added to the hoc template"""
    param_locations = defaultdict(list)
    global_params = {}
    for param in parameters:
        if isinstance(param, NrnGlobalParameter):
            global_params[param.param_name] = param.value
        elif isinstance(param, MetaParameter):
            pass
        else:
            assert isinstance(
                param.locations, (tuple, list)), 'Must have locations list'
            for location in param.locations:
                locs = loc_desc(location, param)
                if not isinstance(locs, list):
                    param_locations[locs].append(param)
                else:
                    for loc in locs:
                        param_locations[loc].append(param)

    section_params = defaultdict(list)
    pprocess_params = defaultdict(list)
    range_params = []

    for loc in param_locations:
        if loc not in location_order:
            location_order.append(loc)

    for loc in location_order:
        if loc not in param_locations:
            continue
        for param in param_locations[loc]:
            if not isinstance(param.param_dependencies, list) or \
                    len(param.param_dependencies) > 0:
                raise CreateHocException(  # also an ACC exception
                    'Exporting models with parameters that have'
                    ' param_dependencies is not yet supported.')
            if isinstance(param, NrnRangeParameter):
                if isinstance(
                        param.value_scaler,
                        NrnSegmentSomaDistanceScaler):
                    range_params.append(
                        RangeExpr(loc,
                                  param.param_name,
                                  param.value,
                                  param.value_scaler))
                elif isinstance(param.value_scaler, NrnSegmentLinearScaler):
                    value = param.value_scale_func(param.value)
                    section_params[loc].append(
                        Location(param.param_name, format_float(value)))
            elif isinstance(param, NrnSectionParameter):
                value = param.value_scale_func(param.value)
                section_params[loc].append(
                    Location(param.param_name, format_float(value)))
            elif isinstance(param, NrnPointProcessParameter):
                value = param.value
                pprocess_params[loc].append(
                    PointExpr(param.param_name, param.locations,
                              format_float(value)))

    ordered_section_params = [(loc, section_params[loc])
                              for loc in location_order]

    ordered_pprocess_params = [(loc, pprocess_params[loc])
                               for loc in location_order]

    return global_params, ordered_section_params, range_params, \
        ordered_pprocess_params, location_order


def _read_template(template_dir, template_filename):
    """Read Jinja2 hoc template to render"""
    if template_dir is None:
        template_dir = os.path.abspath(
            os.path.join(
                os.path.dirname(__file__),
                'templates'))

    template_path = os.path.join(template_dir, template_filename)
    with open(template_path) as template_file:
        template = template_file.read()
        template = jinja2.Template(template)
    return template


def _get_template_params(
        mechs,
        parameters,
        ignored_globals,
        disable_banner,
        default_location_order,
        loc_desc):
    '''return parameters to render Jinja2 templates with simulator descriptions

    Args:
        mechs (): All the mechs for the hoc template
        parameters (): All the parameters in the hoc template
        ignored_globals (iterable str): HOC coded is added for each
        NrnGlobalParameter
        that exists, to test that it matches the values set in the parameters.
        This iterable contains parameter names that aren't checked
        default_location_order (): list of ordered simulator-specific locations
        to use by default
        loc_desc (): method that extracts simulator-specific location
        description from pair of locations and mechanisms/parameters
    '''

    global_params, section_params, range_params, \
        pprocess_params, location_order = \
        _generate_parameters(parameters, default_location_order, loc_desc)

    channels, point_channels = _generate_channels_by_location(
        mechs, location_order, loc_desc)

    ignored_global_params = {}
    for ignored_global in ignored_globals:
        if ignored_global in global_params:
            ignored_global_params[
                ignored_global] = global_params[ignored_global]
            del global_params[ignored_global]

    if not disable_banner:
        banner = 'Created by BluePyOpt(%s) at %s' % (
            bluepyopt.__version__, datetime.now())
    else:
        banner = None

    return dict(global_params=global_params,
                ignored_global_params=ignored_global_params,
                section_params=section_params,
                range_params=range_params,
                pprocess_params=pprocess_params,
                location_order=location_order,
                channels=channels,
                point_channels=point_channels,
                banner=banner)


def create_hoc(mechs,
               parameters,
               morphology=None,
               ignored_globals=(),
               replace_axon=None,
               template_name='CCell',
               template_filename='cell_template.jinja2',
               disable_banner=None,
               template_dir=None,
               custom_jinja_params=None):
    '''return a string containing the hoc template

    Args:
        mechs (): All the mechs for the hoc template
        parameters (): All the parameters in the hoc template
        morpholgy (str): Name of morphology
        ignored_globals (iterable str): HOC coded is added for each
        NrnGlobalParameter
        that exists, to test that it matches the values set in the parameters.
        This iterable contains parameter names that aren't checked
        replace_axon (str): String replacement for the 'replace_axon' command.
        Must include 'proc replace_axon(){ ... }
        template_filename (str): file name of the jinja2 template
        template_dir (str): dir name of the jinja2 template
        custom_jinja_params (dict): dict of additional jinja2 params in case
        of a custom template
    '''

    template = _read_template(template_dir, template_filename)

    template_params = _get_template_params(mechs,
                                           parameters,
                                           ignored_globals,
                                           disable_banner,
                                           DEFAULT_LOCATION_ORDER,
                                           _loc_desc)

    # delete empty dicts to avoid conflict with custom_jinja_params
    del template_params['pprocess_params']
    del template_params['point_channels']

    template_params['range_params'] = range_exprs_to_hoc(
        template_params['range_params']
    )
    re_init_rng = generate_reinitrng(mechs)

    if custom_jinja_params is None:
        custom_jinja_params = {}

    return template.render(template_name=template_name,
                           morphology=morphology,
                           replace_axon=replace_axon,
                           re_init_rng=re_init_rng,
                           **template_params,
                           **custom_jinja_params)


class CreateHocException(Exception):

    """All exceptions generated by create_hoc module"""

    def __init__(self, message):
        """Constructor"""

        super(CreateHocException, self).__init__(message)



"""eFeature classes"""

"""
Copyright (c) 2016-2020, EPFL/Blue Brain Project

 This file is part of BluePyOpt <https://github.com/BlueBrain/BluePyOpt>

 This library is free software; you can redistribute it and/or modify it under
 the terms of the GNU Lesser General Public License version 3.0 as published
 by the Free Software Foundation.

 This library is distributed in the hope that it will be useful, but WITHOUT
 ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
 FOR A PARTICULAR PURPOSE.  See the GNU Lesser General Public License for more
 details.

 You should have received a copy of the GNU Lesser General Public License
 along with this library; if not, write to the Free Software Foundation, Inc.,
 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.
"""

# pylint: disable=R0914

import logging
import numpy as np

from bluepyopt.ephys.base import BaseEPhys
from bluepyopt.ephys.serializer import DictMixin
from .extra_features_utils import *

logger = logging.getLogger(__name__)


def masked_cosine_distance(exp, model):
    from scipy.spatial import distance

    exp_mask = np.isfinite(exp)
    model_mask = np.isfinite(model)
    valid_mask = exp_mask & model_mask

    score = distance.cosine(
        exp[valid_mask], model[valid_mask]
    )

    score *= sum(exp_mask) / len(valid_mask)

    return score


class EFeature(BaseEPhys):

    """EPhys feature"""
    pass


class eFELFeature(EFeature, DictMixin):

    """eFEL feature"""

    SERIALIZED_FIELDS = ('name', 'efel_feature_name', 'recording_names',
                         'stim_start', 'stim_end', 'exp_mean',
                         'exp_std', 'threshold', 'comment')

    def __init__(
            self,
            name,
            efel_feature_name=None,
            recording_names=None,
            stim_start=None,
            stim_end=None,
            exp_mean=None,
            exp_std=None,
            threshold=None,
            stimulus_current=None,
            comment='',
            interp_step=None,
            double_settings=None,
            int_settings=None,
            string_settings=None,
            force_max_score=False,
            max_score=250
    ):
        """Constructor

        Args:
            name (str): name of the eFELFeature object
            efel_feature_name (str): name of the eFeature in the eFEL library
                (ex: 'AP1_peak')
            recording_names (dict): eFEL features can accept several recordings
                as input
            stim_start (float): stimulation start time (ms)
            stim_end (float): stimulation end time (ms)
            exp_mean (float): experimental mean of this eFeature
            exp_std(float): experimental standard deviation of this eFeature
            threshold(float): spike detection threshold (mV)
            comment (str): comment
            interp_step(float): interpolation step (ms)
            double_settings(dict): dictionary with efel double settings that
                should be set before extracting the features
            int_settings(dict): dictionary with efel int settings that
                should be set before extracting the features
            string_settings(dict): dictionary with efel string settings that
                should be set before extracting the features
        """

        super(eFELFeature, self).__init__(name, comment)

        self.recording_names = recording_names
        self.efel_feature_name = efel_feature_name
        self.exp_mean = exp_mean
        self.exp_std = exp_std
        self.stim_start = stim_start
        self.stim_end = stim_end
        self.threshold = threshold
        self.interp_step = interp_step
        self.stimulus_current = stimulus_current
        self.double_settings = double_settings
        self.int_settings = int_settings
        self.string_settings = string_settings
        self.force_max_score = force_max_score
        self.max_score = max_score

    def _construct_efel_trace(self, responses):
        """Construct trace that can be passed to eFEL"""

        trace = {}
        if '' not in self.recording_names:
            raise Exception(
                'eFELFeature: \'\' needs to be in recording_names')
        for location_name, recording_name in self.recording_names.items():
            if location_name == '':
                postfix = ''
            else:
                postfix = ';%s' % location_name

            if recording_name not in responses:
                logger.debug(
                    "Recording named %s not found in responses %s",
                    recording_name,
                    str(responses))
                return None

            if responses[self.recording_names['']] is None or \
                    responses[recording_name] is None:
                return None
            trace['T%s' % postfix] = \
                responses[self.recording_names['']]['time']
            trace['V%s' % postfix] = responses[recording_name]['voltage']
            trace['stim_start%s' % postfix] = [self.stim_start]
            trace['stim_end%s' % postfix] = [self.stim_end]

        return trace

    def _setup_efel(self):
        """Set up efel before extracting the feature"""

        import efel
        efel.reset()

        if self.threshold is not None:
            efel.setThreshold(self.threshold)

        if self.stimulus_current is not None:
            efel.setDoubleSetting('stimulus_current', self.stimulus_current)

        if self.interp_step is not None:
            efel.setDoubleSetting('interp_step', self.interp_step)

        if self.double_settings is not None:
            for setting_name, setting_value in self.double_settings.items():
                efel.setDoubleSetting(setting_name, setting_value)

        if self.int_settings is not None:
            for setting_name, setting_value in self.int_settings.items():
                efel.setIntSetting(setting_name, setting_value)

        if self.string_settings is not None:
            for setting_name, setting_value in self.string_settings.items():
                efel.setStrSetting(setting_name, setting_value)

    def calculate_feature(self, responses, raise_warnings=False):
        """Calculate feature value"""

        efel_trace = self._construct_efel_trace(responses)

        if efel_trace is None:
            feature_value = None
        else:
            self._setup_efel()

            import efel
            values = efel.getMeanFeatureValues(
                [efel_trace],
                [self.efel_feature_name],
                raise_warnings=raise_warnings)
            feature_value = values[0][self.efel_feature_name]

            efel.reset()

        logger.debug(
            'Calculated value for %s: %s',
            self.name,
            str(feature_value))

        return feature_value

    def calculate_score(self, responses, trace_check=False):
        """Calculate the score"""

        efel_trace = self._construct_efel_trace(responses)

        if efel_trace is None:
            score = self.max_score
        else:
            self._setup_efel()

            import efel
            score = efel.getDistance(
                efel_trace,
                self.efel_feature_name,
                self.exp_mean,
                self.exp_std,
                trace_check=trace_check,
                error_dist=self.max_score
            )
            if self.force_max_score:
                score = min(score, self.max_score)

            efel.reset()

        logger.debug('Calculated score for %s: %f', self.name, score)

        return score

    def __str__(self):
        """String representation"""

        return "%s for %s with stim start %s and end %s, " \
            "exp mean %s and std %s and AP threshold override %s" % \
            (self.efel_feature_name,
             self.recording_names,
             self.stim_start,
             self.stim_end,
             self.exp_mean,
             self.exp_std,
             self.threshold)


class extraFELFeature(EFeature, DictMixin):
    """extraFEL feature"""

    SERIALIZED_FIELDS = ('name', 'extrafel_feature_name', 'recording_names',
                         'somatic_recording_name', 'fcut', 'fs',
                         'channel_ids', 'stim_start', 'stim_end',
                         'exp_mean', 'exp_std', 'threshold', 'comment')

    def __init__(
            self,
            name,
            extrafel_feature_name=None,
            recording_names=None,
            somatic_recording_name=None,
            fcut=None,
            fs=None,
            filt_type=None,
            ms_cut=None,
            upsample=None,
            skip_first_spike=True,
            skip_last_spike=True,
            channel_ids=None,
            stim_start=None,
            stim_end=None,
            exp_mean=None,
            exp_std=None,
            threshold=None,
            comment='',
            interp_step=None,
            double_settings=None,
            int_settings=None,
            force_max_score=False,
            max_score=250,
    ):
        """Constructor

        Args:
            name (str): name of the extraFELFeature object
            extrafel_feature_name (str): name of the eFeature in the
                spikefeatures library (ex: 'halfwidth')
            recording_names (dict): eFEL features can accept several
                recordings as input
            somatic_recording_name (str): intracellualar recording from soma,
                used to detect spikes. If None, spikes are detected from
                extracellular trace
            fcut (float, array, or None): cutoff frequency(ies) for filter. If
                float, a high-pass filter is used. If array-like a bandpass
                filter is used. If None, traces are note filtered
            fs (float): sampling frequency to resample extracellular traces
                (in kHz)
            filt_type (str): type of the bandpass filter used
                (default 'filtfilt')
            ms_cut (float, list, or None): cut in ms before and after the
                intra peak. If scalar, the cut is symmetrical
            upsample (int, or None): upsample factor for average waveform
                before computing features
            skip_first_spike (bool): if True, the first spike is skipped
                before computing the average waveform
                (to avoid artifacts)
            skip_last_spike (bool): if True, the last spike is skipped
                before computing the average waveform
                (to avoid artifacts)
            channel_ids (int, np.array, or None): if None, all channels are
                used to compute the feature and calculate the score
                (using the cosine_dist). If int, a single channel is used and
                the score is the normalised deviation form the exp value.
                If list/array, the cosine distance is computed over a subset
                of channels
            stim_start (float): stimulation start time (ms)
            stim_end (float): stimulation end time (ms)
            exp_mean (list of floats): experimental mean of this eFeature
            exp_std (list of floats): experimental standard deviation
                of this eFeature
            threshold (float): spike detection threshold (mV)
            comment (str): comment
            interp_step (float): interpolation step (ms)
            double_settings (dict): dictionary with efel double settings that
                should be set before extracting the features
            int_settings (dict): dictionary with efel int settings that
                should be set before extracting the features
        """

        super(extraFELFeature, self).__init__(name, comment)

        self.recording_names = recording_names
        self.somatic_recording_name = somatic_recording_name
        self.extrafel_feature_name = extrafel_feature_name
        self.fcut = fcut
        self.fs = fs
        self.filt_type = filt_type
        self.ms_cut = ms_cut
        self.upsample = upsample
        self.skip_first_spike = skip_first_spike
        self.skip_last_spike = skip_last_spike
        self.channel_ids = channel_ids
        self.exp_mean = exp_mean
        self.exp_std = exp_std
        self.stim_start = stim_start
        self.stim_end = stim_end
        self.threshold = threshold
        self.interp_step = interp_step
        self.double_settings = double_settings
        self.int_settings = int_settings
        self.force_max_score = force_max_score
        self.max_score = max_score

    def _construct_somatic_efel_trace(self, responses):
        """Construct trace that can be passed to eFEL"""

        trace = {}
        if self.somatic_recording_name not in responses:
            logger.debug(
                "Recording named %s not found in responses %s",
                self.somatic_recording_name,
                str(responses),
            )
            return None

        if responses[self.somatic_recording_name] is None:
            return None

        response = responses[self.somatic_recording_name]

        trace["T"] = response["time"]
        trace["V"] = response["voltage"]
        trace["stim_start"] = [self.stim_start]
        trace["stim_end"] = [self.stim_end]

        return trace

    def _setup_efel(self):
        """Set up efel before extracting the feature"""

        import efel

        efel.reset()

        if self.threshold is not None:
            efel.setThreshold(self.threshold)

        if self.interp_step is not None:
            efel.setDoubleSetting("interp_step", self.interp_step)

        if self.double_settings is not None:
            for setting_name, setting_value in self.double_settings.items():
                efel.setDoubleSetting(setting_name, setting_value)

        if self.int_settings is not None:
            for setting_name, setting_value in self.int_settings.items():
                efel.setIntSetting(setting_name, setting_value)

    def _get_peak_times(self, responses, raise_warnings=False):

        efel_trace = self._construct_somatic_efel_trace(responses)

        if efel_trace is None:
            peak_times = None
        else:
            self._setup_efel()

            import efel

            peaks = efel.getFeatureValues(
                [efel_trace], ["peak_time"], raise_warnings=raise_warnings
            )
            peak_times = peaks[0]["peak_time"]

            efel.reset()

        return peak_times

    def calculate_feature(
            self,
            responses,
            raise_warnings=False,
            return_waveforms=False,
    ):
        from .extra_features_utils import calculate_features

        """Calculate feature value"""
        peak_times = self._get_peak_times(
            responses, raise_warnings=raise_warnings
        )

        if len(peak_times) > 1 and self.skip_first_spike:
            peak_times = peak_times[1:]

        if len(peak_times) > 1 and self.skip_last_spike:
            peak_times = peak_times[:-1]

        if responses[self.recording_names[""]] is not None:
            response = responses[self.recording_names[""]]
        else:
            return None

        if np.std(np.diff(response["time"])) > 0.001 * np.mean(
                np.diff(response["time"])
        ):
            assert self.fs is not None
            logger.info("extraFELFeature.calculate_feature: interpolate")
            response_interp = _interpolate_response(response, fs=self.fs)
        else:
            response_interp = response

        if self.fcut is not None:
            logger.info("extraFELFeature.calculate_feature: enabled")
            response_filter = _filter_response(response_interp,
                                               fcut=self.fcut,
                                               filt_type=self.filt_type)
        else:
            logger.info("extraFELFeature.calculate_feature: filter disabled")
            response_filter = response_interp

        ewf = _get_waveforms(response_filter, peak_times, self.ms_cut)
        mean_wf = np.mean(ewf, axis=0)

        values = calculate_features(
            mean_wf,
            self.fs * 1000,
            upsample=self.upsample,
            feature_names=[self.extrafel_feature_name]
        )

        feature_value = values[self.extrafel_feature_name]

        if self.channel_ids is not None:
            feature_value = feature_value[self.channel_ids]

        logger.debug(
            "Calculated value for %s: %s", self.name, str(feature_value)
        )

        if return_waveforms:
            return feature_value, mean_wf
        else:
            return feature_value

    def calculate_score(self, responses, trace_check=False):
        """Calculate the score"""

        if (
                responses[self.recording_names[""].replace("soma.v",
                                                           "MEA.LFP")]
                is None
                or responses[self.recording_names[""]] is None
        ):
            return self.max_score

        feature_value = self.calculate_feature(responses)

        if np.isscalar(feature_value):
            # scalar feature
            if np.isfinite(feature_value):
                score = np.abs((feature_value - self.exp_mean)) / self.exp_std
            else:
                score = self.max_score
            if not np.isfinite(score):
                logger.debug(
                    f"Found score nan value {self.extrafel_feature_name} "
                    f"- std: {self.exp_std} - channel: {self.channel_ids}"
                )
                score = self.max_score
        else:
            score = masked_cosine_distance(
                np.asarray(self.exp_mean),
                np.asarray(feature_value)
            )

        if np.isnan(score):
            score = self.max_score

        if self.force_max_score:
            score = min(score, self.max_score)

        logger.debug("Calculated score for %s: %f", self.name, score)

        return score

    def __str__(self):
        """String representation"""

        return ("%s for %s with stim start %s and end %s, "
                "exp mean %s and std %s and AP threshold override %s"
                % (self.extrafel_feature_name,
                   self.recording_names,
                   self.stim_start,
                   self.stim_end,
                   self.exp_mean,
                   self.exp_std,
                   self.threshold)
                )


def _interpolate_response(response, fs=20.0):
    from scipy.interpolate import interp1d

    x = response["time"]
    y = response["voltage"]
    f = interp1d(x, y, axis=1)
    xnew = np.arange(np.min(x), np.max(x), 1.0 / fs)
    ynew = f(xnew)  # use interpolation function returned by `interp1d`

    response_new = {}
    response_new["time"] = xnew
    response_new["voltage"] = ynew

    return response_new


def _filter_response(response, fcut=[0.5, 6000], order=2, filt_type="lfilter"):
    import scipy.signal as ss

    fs = 1 / np.mean(np.diff(response["time"])) * 1000
    fn = fs / 2.0

    trace = response["voltage"]

    if isinstance(fcut, (float, int, np.floating, np.integer)):
        btype = "highpass"
        band = fcut / fn
    else:
        assert isinstance(fcut, (list, np.ndarray)) and len(fcut) == 2
        btype = "bandpass"
        band = np.array(fcut) / fn

    b, a = ss.butter(order, band, btype=btype)

    if len(trace.shape) == 2:
        if filt_type == "filtfilt":
            filtered = ss.filtfilt(b, a, trace, axis=1)
        else:
            filtered = ss.lfilter(b, a, trace, axis=1)
    else:
        if filt_type == "filtfilt":
            filtered = ss.filtfilt(b, a, trace)
        else:
            filtered = ss.lfilter(b, a, trace)

    response_new = {}
    response_new["time"] = response["time"]
    response_new["voltage"] = filtered

    return response_new


def _get_waveforms(response, peak_times, snippet_len_ms):
    times = response["time"]
    traces = response["voltage"]

    assert np.std(np.diff(times)) < 0.001 * np.mean(
        np.diff(times)
    ), "Sampling frequency must be constant"

    fs = 1.0 / np.mean(np.diff(times))  # kHz

    reference_frames = (peak_times * fs).astype(int)

    if isinstance(snippet_len_ms, (tuple, list, np.ndarray)):
        snippet_len_before = int(snippet_len_ms[0] * fs)
        snippet_len_after = int(snippet_len_ms[1] * fs)
    else:
        snippet_len_before = int((snippet_len_ms + 1) / 2 * fs)
        snippet_len_after = int((snippet_len_ms - snippet_len_before) * fs)

    num_snippets = len(peak_times)
    if len(traces.shape) == 2:
        num_channels = traces.shape[0]
    else:
        num_channels = 1
        traces = traces[np.newaxis, :]
    num_frames = len(times)
    snippet_len_total = int(snippet_len_before + snippet_len_after)
    waveforms = np.zeros(
        (num_snippets, num_channels, snippet_len_total), dtype=traces.dtype
    )

    for i in range(num_snippets):
        snippet_chunk = np.zeros(
            (num_channels, snippet_len_total), dtype=traces.dtype
        )
        if 0 <= reference_frames[i] < num_frames:
            snippet_range = np.array(
                [
                    int(reference_frames[i]) - snippet_len_before,
                    int(reference_frames[i]) + snippet_len_after,
                ]
            )
            snippet_buffer = np.array([0, snippet_len_total], dtype="int")
            # The following handles the out-of-bounds cases
            if snippet_range[0] < 0:
                snippet_buffer[0] -= snippet_range[0]
                snippet_range[0] -= snippet_range[0]
            if snippet_range[1] >= num_frames:
                snippet_buffer[1] -= snippet_range[1] - num_frames
                snippet_range[1] -= snippet_range[1] - num_frames
            snippet_chunk[:, snippet_buffer[0]:snippet_buffer[1]] = \
                traces[:, snippet_range[0]:snippet_range[1]]
        waveforms[i] = snippet_chunk

    return waveforms



'''
Note: this is a bizarre test due to the fact that scoop can't be started from
within python:
https://github.com/soravux/scoop/issues/29

Therefore, the way this works is for the test_ that's detected by nt then
creates a subprocess that runs scoop using the normal command line arguments

It then captures the output, and looks for the BEST: magic string which
should match the precomputed output
'''

import os

import subprocess

import bluepyopt as nrp
import bluepyopt.ephys as nrpel

SIMPLE_SWC = os.path.join(os.path.abspath(os.path.dirname(__file__)),
                          '../../examples/simplecell/simple.swc')


# Disabled this test. Doesn't work on a mac for some reason
# Error message: can't import scoop
# TODO Renable once this is fixed
def disabled_scoop():
    """Simplecell: test scoop"""
    cmd = ['python', '-m', 'scoop', '-n', '2', __file__]
    output = subprocess.check_output(cmd)
    for line in output.split('\n'):
        if line.startswith('BEST'):
            break
    assert line == 'BEST: [0.11268238279399023, 0.038129859413828474]'


# The rest defines the optimization we run with scoop

morph = nrpel.morphologies.NrnFileMorphology(SIMPLE_SWC)
somatic_loc = nrpel.locations.NrnSeclistLocation('somatic',
                                                 seclist_name='somatic')

hh_mech = nrpel.mechanisms.NrnMODMechanism(name='hh',
                                           suffix='hh',
                                           locations=[somatic_loc])

cm_param = nrpel.parameters.NrnSectionParameter(name='cm',
                                                param_name='cm',
                                                value=1.0,
                                                locations=[somatic_loc],
                                                frozen=True)


gnabar_param = nrpel.parameters.NrnSectionParameter(name='gnabar_hh',
                                                    param_name='gnabar_hh',
                                                    locations=[somatic_loc],
                                                    bounds=[0.05, 0.125],
                                                    frozen=False)

gkbar_param = nrpel.parameters.NrnSectionParameter(name='gkbar_hh',
                                                   param_name='gkbar_hh',
                                                   bounds=[0.01, 0.075],
                                                   locations=[somatic_loc],
                                                   frozen=False)

simple_cell = nrpel.celltemplate.CellTemplate(name='simple_cell',
                                              morph=morph,
                                              mechs=[hh_mech],
                                              params=[cm_param,
                                                      gnabar_param,
                                                      gkbar_param])

soma_loc = nrpel.locations.NrnSeclistCompLocation(name='soma',
                                                  seclist_name='somatic',
                                                  sec_index=0,
                                                  comp_x=0.5)

protocols = {}
for protocol_name, amplitude in [('step1', 0.01), ('step2', 0.05)]:
    stim = nrpel.stimuli.NrnSquarePulse(step_amplitude=amplitude,
                                        step_delay=100,
                                        step_duration=50,
                                        location=soma_loc,
                                        total_duration=200)
    rec = nrpel.recordings.CompRecording(name='%s.soma.v' % protocol_name,
                                         location=soma_loc,
                                         variable='v')
    protocol = nrpel.protocols.Protocol(protocol_name, [stim], [rec])
    protocols[protocol.name] = protocol


default_params = {'gnabar_hh': 0.1, 'gkbar_hh': 0.03}
responses = simple_cell.run_protocols(protocols,
                                      param_values=default_params)

efel_feature_means = {'step1': {'Spikecount': 1}, 'step2': {'Spikecount': 5}}

objectives = []

for protocol_name, protocol in protocols.items():
    stim_start = protocol.stimuli[0].step_delay
    stim_end = stim_start + protocol.stimuli[0].step_duration
    for efel_feature_name, mean in \
            efel_feature_means[protocol_name].items():
        feature_name = '%s.%s' % (protocol_name, efel_feature_name)
        feature = nrpel.efeatures.eFELFeature(
            feature_name,
            efel_feature_name=efel_feature_name,
            recording_names={'': '%s.soma.v' % protocol_name},
            stim_start=stim_start,
            stim_end=stim_end,
            exp_mean=mean,
            exp_std=0.05 * mean)
        objective = objective = nrpel.objectives.SingletonObjective(
            feature_name,
            feature)
        objectives.append(objective)


score_calc = nrpel.scorecalculators.ObjectivesScoreCalculator(objectives)
cell_evaluator = nrpel.cellevaluator.CellEvaluator(
    cell_template=simple_cell,
    param_names=[
        'gnabar_hh',
        'gkbar_hh'],
    fitness_protocols=protocols,
    fitness_calculator=score_calc)

optimisation = nrp.Optimisation(
    evaluator=cell_evaluator,
    eval_function=cell_evaluator.evaluate_with_lists,
    offspring_size=10,
    use_scoop=True)

if __name__ == '__main__':
    final_pop, hall_of_fame, logs, hist = optimisation.run(max_ngen=2)
    print('BEST:', hall_of_fame[0])



"""Utils function"""

"""
Copyright (c) 2016-2022, EPFL/Blue Brain Project

 This file is part of BluePyOpt <https://github.com/BlueBrain/BluePyOpt>

 This library is free software; you can redistribute it and/or modify it under
 the terms of the GNU Lesser General Public License version 3.0 as published
 by the Free Software Foundation.

 This library is distributed in the hope that it will be useful, but WITHOUT
 ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
 FOR A PARTICULAR PURPOSE.  See the GNU Lesser General Public License for more
 details.

 You should have received a copy of the GNU Lesser General Public License
 along with this library; if not, write to the Free Software Foundation, Inc.,
 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.
"""

import numpy
import random

import deap.base

# pylint: disable=R0914, R0912


class WeightedReducedFitness(deap.base.Fitness):

    """Fitness that compares by weighted objective values"""

    def __init__(self, values=(), obj_size=None, reduce_fcn=numpy.sum):
        self.weights = [-1.0] * obj_size if obj_size is not None else [-1]
        self.reduce_fcn = reduce_fcn

        super(WeightedReducedFitness, self).__init__(values)

    @property
    def reduce(self):
        """Reduce of values"""
        return self.reduce_fcn(self.values)

    @property
    def weighted_reduce(self):
        """Reduce of weighted values"""
        return self.reduce_fcn(self.wvalues)

    def __le__(self, other):
        return self.weighted_reduce <= other.weighted_reduce

    def __lt__(self, other):
        return self.weighted_reduce < other.weighted_reduce

    def __deepcopy__(self, _):
        """Override deepcopy"""

        cls = self.__class__
        result = cls.__new__(cls)
        result.__dict__.update(self.__dict__)
        return result


class WSListIndividual(list):

    """Individual consisting of a list with weighted fitness"""

    def __init__(self, *args, **kwargs):
        """Constructor"""

        reduce_fcn = kwargs.get("reduce_fcn", numpy.sum)
        self.fitness = WeightedReducedFitness(
            obj_size=kwargs["obj_size"], reduce_fcn=reduce_fcn
        )

        # Index of the parent, used by MO-CMA
        self._ps = "p", 0

        del kwargs["obj_size"]
        if "reduce_fcn" in kwargs:
            del kwargs["reduce_fcn"]
        super(WSListIndividual, self).__init__(*args, **kwargs)


def update_history_and_hof(halloffame, history, population):
    """Update the hall of fame with the generated individuals
    Note: History and Hall-of-Fame behave like dictionaries
    """
    if halloffame is not None:
        halloffame.update(population)

    history.update(population)


def record_stats(stats, logbook, gen, population, invalid_count):
    """Update the statistics with the new population"""
    record = stats.compile(population) if stats is not None else {}
    logbook.record(gen=gen, nevals=invalid_count, **record)


def closest_feasible(individual, lbounds, ubounds):
    """Returns the closest individual in the parameter bounds"""
    # TODO: Fix 1e-9 hack
    for i, (u, l, el) in enumerate(zip(ubounds, lbounds, individual)):
        if el >= u:
            individual[i] = u - 1e-9
        elif el <= l:
            individual[i] = l + 1e-9
    return individual


def bound(population, lbounds, ubounds):
    """Bounds the population based on lower and upper parameter bounds."""
    n_out = 0
    for i, ind in enumerate(population):
        if numpy.any(numpy.less(ind, lbounds)) or numpy.any(
            numpy.greater(ind, ubounds)
        ):
            population[i] = closest_feasible(ind, lbounds, ubounds)
            n_out += 1
    return n_out


def uniform(lower_list, upper_list, dimensions):
    """Uniformly pick an individual"""
    if hasattr(lower_list, "__iter__"):
        return [
            random.uniform(lower, upper) for lower, upper in
            zip(lower_list, upper_list)
        ]
    else:
        return [random.uniform(lower_list, upper_list) for _ in
                range(dimensions)]


def reduce_method(meth):
    """Overwrite reduce"""
    return (getattr, (meth.__self__, meth.__func__.__name__))


def run_next_gen(criteria, terminator):
    """Condition to stay inside the loop."""
    if terminator is None:
        return criteria
    return criteria and not terminator.is_set()



"""Tests for ephys.mechanisms"""

import string
import random
import json
import difflib


import pytest

from . import utils
from bluepyopt import ephys
import bluepyopt.ephys.examples.simplecell
simplecell = bluepyopt.ephys.examples.simplecell.SimpleCell()

from bluepyopt.ephys.serializer import instantiator

simple_cell = simplecell.cell_model
simple_cell.freeze(simplecell.default_param_values)
sim = simplecell.nrn_sim


@pytest.mark.unit
def test_mechanism_serialize():
    """ephys.mechanisms: Testing serialize"""
    mech = utils.make_mech()
    serialized = mech.to_dict()
    assert isinstance(json.dumps(serialized), str)
    deserialized = instantiator(serialized)
    assert isinstance(deserialized, ephys.mechanisms.NrnMODMechanism)


@pytest.mark.unit
def test_nrnmod_instantiate():
    """ephys.mechanisms: Testing insert mechanism"""

    test_mech = ephys.mechanisms.NrnMODMechanism(
        'test.pas',
        suffix='pas',
        locations=[simplecell.somatic_loc])

    assert str(test_mech) == "test.pas: pas at ['somatic']"

    simple_cell.instantiate(sim=sim)

    test_mech.instantiate(sim=sim, icell=simple_cell.icell)
    test_mech.destroy(sim=sim)

    simple_cell.destroy(sim=sim)

    pytest.raises(TypeError, ephys.mechanisms.NrnMODMechanism,
                  'test.pas',
                  suffix='pas',
                  prefix='pas',
                  locations=[simplecell.somatic_loc])

    test_mech = ephys.mechanisms.NrnMODMechanism(
        'test.pas',
        prefix='pas',
        locations=[simplecell.somatic_loc])

    assert test_mech.suffix == 'pas'

    test_mech.prefix = 'pas2'
    assert test_mech.suffix == 'pas2'

    test_mech = ephys.mechanisms.NrnMODMechanism(
        'unknown',
        suffix='unknown',
        locations=[simplecell.somatic_loc])

    simple_cell.instantiate(sim=sim)

    pytest.raises(
        ValueError,
        test_mech.instantiate,
        sim=sim,
        icell=simple_cell.icell)

    test_mech.destroy(sim=sim)
    simple_cell.destroy(sim=sim)


def compare_strings(s1, s2):
    """Compare two strings"""

    diff = list(difflib.unified_diff(s1.splitlines(1), s2.splitlines(1)))

    if len(diff) > 0:
        print(''.join(diff))
        return False
    else:
        return True


@pytest.mark.unit
def test_nrnmod_reinitrng_block():
    """ephys.mechanisms: Testing reinitrng_block"""

    test_mech = ephys.mechanisms.NrnMODMechanism(
        'stoch',
        suffix='Stoch',
        locations=[simplecell.somatic_loc])

    block = test_mech.generate_reinitrng_hoc_block()
    expected_block = '    forsec somatic { deterministic_Stoch = 1 }\n'

    assert compare_strings(block, expected_block)

    test_mech = ephys.mechanisms.NrnMODMechanism(
        'stoch',
        suffix='Stoch',
        deterministic=False,
        locations=[simplecell.somatic_loc])

    block = test_mech.generate_reinitrng_hoc_block()

    expected_block = \
        """    forsec somatic {
        for (x, 0) {
            setdata_Stoch(x)
            sf.tail(secname(), "\\\\.", name)
            sprint(full_str, "%s.%.19g", name, x)
            if (channel_seed_set) {
                setRNG_Stoch(gid, hash_str(full_str), channel_seed)
            } else {
                setRNG_Stoch(gid, hash_str(full_str))
            }
        }
    }
"""

    assert compare_strings(block, expected_block)


@pytest.mark.unit
def test_nrnmod_determinism():
    """ephys.mechanisms: Testing determinism"""

    test_mech = ephys.mechanisms.NrnMODMechanism(
        'pas',
        suffix='pas',
        deterministic=False,
        locations=[simplecell.somatic_loc])

    simple_cell.instantiate(sim=sim)

    pytest.raises(
        TypeError,
        test_mech.instantiate,
        sim=sim,
        icell=simple_cell.icell)
    test_mech.destroy(sim=sim)

    simple_cell.destroy(sim=sim)


@pytest.mark.unit
def test_pprocess_instantiate():
    """ephys.mechanisms: Testing insert point process"""

    test_pprocess = ephys.mechanisms.NrnMODPointProcessMechanism(
        name='expsyn',
        suffix='ExpSyn',
        locations=[simplecell.somacenter_loc])

    assert (
        str(test_pprocess) ==
        "expsyn: ExpSyn at ['somatic[0](0.5)']")

    simple_cell.instantiate(sim=sim)

    assert test_pprocess.pprocesses is None

    test_pprocess.instantiate(sim=sim, icell=simple_cell.icell)
    assert len(test_pprocess.pprocesses) == 1
    pprocess = test_pprocess.pprocesses[0]

    assert hasattr(pprocess, 'tau')
    test_pprocess.destroy(sim=sim)

    assert test_pprocess.pprocesses is None

    simple_cell.destroy(sim=sim)

    test_pprocess = ephys.mechanisms.NrnMODPointProcessMechanism(
        name='expsyn',
        suffix='Exp',
        locations=[simplecell.somacenter_loc])

    simple_cell.instantiate(sim=sim)

    pytest.raises(
        AttributeError,
        test_pprocess.instantiate,
        sim=sim,
        icell=simple_cell.icell)

    test_pprocess.destroy(sim=sim)
    simple_cell.destroy(sim=sim)


@pytest.mark.unit
def test_string_hash_functions():
    """ephys.mechanisms: Testing string hash function"""

    n_of_strings = 100
    max_size = 50

    random.seed(1)

    test_strings = ['', 'a']
    test_strings += [''.join
                     (random.choice
                      (string.printable)
                      for _ in range(random.choice(range(max_size))))
                     for _ in range(n_of_strings)]
    hashes_py = [
        ephys.mechanisms.NrnMODMechanism.hash_py
        (test_string) for test_string in test_strings]
    hashes_hoc = [
        ephys.mechanisms.NrnMODMechanism.hash_hoc
        (test_string, simplecell.nrn_sim) for test_string in test_strings]

    assert hashes_py == hashes_hoc
    assert hashes_py[:2] == [0.0, 97.0]



"""Parameter scaler classes"""

"""
Copyright (c) 2016-2020, EPFL/Blue Brain Project

 This file is part of BluePyOpt <https://github.com/BlueBrain/BluePyOpt>

 This library is free software; you can redistribute it and/or modify it under
 the terms of the GNU Lesser General Public License version 3.0 as published
 by the Free Software Foundation.

 This library is distributed in the hope that it will be useful, but WITHOUT
 ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
 FOR A PARTICULAR PURPOSE.  See the GNU Lesser General Public License for more
 details.

 You should have received a copy of the GNU Lesser General Public License
 along with this library; if not, write to the Free Software Foundation, Inc.,
 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.
"""

# pylint: disable=W0511

import string

from bluepyopt.ephys.base import BaseEPhys
from bluepyopt.ephys.parameterscalers.acc_iexpr import generate_acc_scale_iexpr
from bluepyopt.ephys.serializer import DictMixin
from bluepyopt.ephys.morphologies import ArbFileMorphology

FLOAT_FORMAT = '%.17g'


def format_float(value):
    """Return formatted float string"""
    return FLOAT_FORMAT % value


class MissingFormatDict(dict):

    """Extend dict for string formatting with missing values"""

    def __missing__(self, key):  # pylint: disable=R0201
        """Return string with format key for missing keys"""
        return '{' + key + '}'


class ParameterScaler(BaseEPhys):

    """Parameter scalers"""
    pass

# TODO get rid of the 'segment' here


class NrnSegmentLinearScaler(ParameterScaler, DictMixin):

    """Linear scaler"""
    SERIALIZED_FIELDS = ('name', 'comment', 'multiplier', 'offset', )

    def __init__(
            self,
            name=None,
            multiplier=1.0,
            offset=0.0,
            comment=''):
        """Constructor

        Args:
            name (str): name of this object
            multiplier (float): slope of the linear scaler
            offset (float): intercept of the linear scaler
        """

        super(NrnSegmentLinearScaler, self).__init__(name, comment)
        self.multiplier = multiplier
        self.offset = offset

    def scale(self, values, segment=None, sim=None):  # pylint: disable=W0613
        """Scale a value based on a segment"""
        if isinstance(values, dict):
            value = values["value"]
        else:
            value = values
        return self.multiplier * value + self.offset

    def __str__(self):
        """String representation"""

        return '%s * value + %s' % (self.multiplier, self.offset)


class NrnSegmentSectionDistanceScaler(ParameterScaler, DictMixin):

    """Scaler based on distance from soma"""
    SERIALIZED_FIELDS = ('name', 'comment', 'distribution',
                         "distribution", "dist_param_names",
                         "ref_sec", "ref_location",)

    def __init__(
            self,
            name=None,
            distribution=None,
            comment='',
            dist_param_names=None,
            ref_section='soma[0]',
            ref_location=0,):
        """Constructor

        Args:
            name (str): name of this object
            distribution (str): distribution of parameter dependent on distance
                from soma. string can contain `distance` and/or `value` as
                placeholders for the distance to the soma and parameter value
                respectivily
            dist_param_names (list): list of names of parameters that
                parametrise the distribution. These names will become
                attributes of this object.
                The distribution string should contain these names, and they
                will be replaced by values of the corresponding attributes
            ref_section (str): string with name of reference section to
                compute distance (e.g. "soma[0]", "dend[2]")
            ref_location (float): location along the soma used as origin
                from which to compute the distances. Expressed as a fraction
                (between 0.0 and 1.0).
        """

        super(NrnSegmentSectionDistanceScaler, self).__init__(name, comment)
        self.distribution = distribution

        self.dist_param_names = dist_param_names
        self.ref_location = ref_location
        self.ref_section = ref_section

        if not (0.0 <= self.ref_location <= 1.0):
            raise ValueError("ref_location must be between 0 and 1.")

        if self.dist_param_names is not None:
            for dist_param_name in self.dist_param_names:
                if dist_param_name not in self.distribution:
                    raise ValueError(
                        'NrnSegmentSectionDistanceScaler: "{%s}" '
                        'missing from distribution string "%s"' %
                        (dist_param_name, distribution))
                setattr(self, dist_param_name, None)

    @property
    def inst_distribution(self):
        """The instantiated distribution"""

        dist_dict = MissingFormatDict()

        if self.dist_param_names is not None:
            for dist_param_name in self.dist_param_names:
                dist_param_value = getattr(self, dist_param_name)
                if dist_param_value is None:
                    raise ValueError("NrnSegmentSomaDistanceScaler: %s "
                                     "was uninitialised" % dist_param_name)
                dist_dict[dist_param_name] = dist_param_value

        # Use this special formatting to bypass missing keys
        return string.Formatter().vformat(self.distribution, (), dist_dict)

    def eval_dist(self, values, distance):
        """Create the final dist string"""

        scale_dict = {}
        if isinstance(values, dict):
            for k, v in values.items():
                scale_dict[k] = format_float(v)
        else:
            scale_dict["value"] = format_float(values)
        scale_dict["distance"] = format_float(distance)

        return self.inst_distribution.format(**scale_dict)

    def scale(self, values, segment, sim=None):
        """Scale a value based on a segment"""

        # find section
        target_sec = None
        for sec in segment.sec.wholetree():
            if "." in sec.name():  # deal with templates
                sec_name = sec.name().split(".")[1]
            else:
                sec_name = sec.name()
            if self.ref_section in sec_name:
                target_sec = sec
                break

        if target_sec is None:
            raise Exception(f"Could not find section {self.ref_section} "
                            f"in section list")

        # Initialise origin
        sim.neuron.h.distance(0, self.ref_location, sec=target_sec)

        distance = sim.neuron.h.distance(1, segment.x, sec=segment.sec)

        # Find something to generalise this
        import math  # pylint:disable=W0611 #NOQA

        # This eval is unsafe (but is it ever dangerous ?)
        # pylint: disable=W0123
        return eval(self.eval_dist(values, distance))

    def acc_scale_iexpr(self, value, constant_formatter=format_float):
        """Generate Arbor scale iexpr for a given value"""
        raise ValueError(
            "Parameter scaling based on general Neuron segment/section"
            " distances not supported in Arbor.")

    def __str__(self):
        """String representation"""

        return self.distribution


class NrnSegmentSomaDistanceScaler(NrnSegmentSectionDistanceScaler,
                                   ParameterScaler, DictMixin):

    """Scaler based on distance from soma"""
    SERIALIZED_FIELDS = ('name', 'comment', 'distribution', )

    def __init__(
            self,
            name=None,
            distribution=None,
            comment='',
            dist_param_names=None,
            soma_ref_location=0.5):
        """Constructor
        Args:
            name (str): name of this object
            distribution (str): distribution of parameter dependent on distance
                from soma. string can contain `distance` and/or `value` as
                placeholders for the distance to the soma and parameter value
                respectivily
            dist_param_names (list): list of names of parameters that
                parametrise the distribution. These names will become
                attributes of this object.
                The distribution string should contain these names, and they
                will be replaced by values of the corresponding attributes
            soma_ref_location (float): location along the soma used as origin
                from which to compute the distances. Expressed as a fraction
                (between 0.0 and 1.0).
        """

        super(NrnSegmentSomaDistanceScaler, self).__init__(
            name, distribution, comment, dist_param_names,
            ref_section='soma[0]', ref_location=soma_ref_location)

    def acc_scale_iexpr(self, value, constant_formatter=format_float):
        """Generate Arbor scale iexpr for a given value"""

        iexpr = self.inst_distribution

        variables = dict(
            value=value,
            distance='(distance %s)' %  # could be a ctor param if required
            ArbFileMorphology.region_labels['somatic'].ref
        )
        return generate_acc_scale_iexpr(iexpr, variables, constant_formatter)



"""Functional LFPy test"""

import os
import sys
import pytest

L5PC_LFPY_PATH = os.path.abspath(
    os.path.join(
        os.path.dirname(__file__),
        '../../examples/l5pc_lfpy'
    )
)
SCRIPT_DIR = os.path.dirname(os.path.abspath(__file__))

sys.path.insert(0, L5PC_LFPY_PATH)

import l5pc_lfpy_evaluator
from generate_extra_features import release_params


@pytest.mark.slow
def test_lfpy_evaluator():
    """Test CellEvaluator with an LFPy cell and LFPy simulator"""

    evaluator = l5pc_lfpy_evaluator.create(
        feature_file=L5PC_LFPY_PATH + "/extra_features.json",
        cvode_active=False,
        dt=0.025,
    )

    responses = evaluator.run_protocols(
        protocols=evaluator.fitness_protocols.values(),
        param_values=release_params
    )
    values = evaluator.fitness_calculator.calculate_values(responses)

    assert len(values) == 21
    assert abs(values['Step1.soma.AP_height'] - 27.85963902931001) < 1e-5
    assert len(responses['Step1.MEA.v']["voltage"]) == 40



"""Multi Objective CMA-es class"""

"""
Copyright (c) 2016-2022, EPFL/Blue Brain Project

 This file is part of BluePyOpt <https://github.com/BlueBrain/BluePyOpt>

 This library is free software; you can redistribute it and/or modify it under
 the terms of the GNU Lesser General Public License version 3.0 as published
 by the Free Software Foundation.

 This library is distributed in the hope that it will be useful, but WITHOUT
 ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
 FOR A PARTICULAR PURPOSE.  See the GNU Lesser General Public License for more
 details.

 You should have received a copy of the GNU Lesser General Public License
 along with this library; if not, write to the Free Software Foundation, Inc.,
 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.
"""

# pylint: disable=R0912, R0914

import logging
import numpy
import copy
from math import log

import deap
from deap import base
from deap import cma

from .stoppingCriteria import MaxNGen, Stagnationv2
from . import utils
from . import hype

logger = logging.getLogger("__main__")


def get_hyped(pop, ubound_score=250., threshold_improvement=240.):
    """Compute the hypervolume contribution of each individual.
    The fitness space is first bounded and all dimension who do not show
    improvement are ignored.
    """

    # Cap the obj at 250
    points = numpy.array([ind.fitness.values for ind in pop])
    points[points > ubound_score] = ubound_score
    lbounds = numpy.min(points, axis=0)
    ubounds = numpy.max(points, axis=0)

    # Remove the dimensions that do not show any improvement
    to_remove = []
    for i, lb in enumerate(lbounds):
        if lb >= threshold_improvement:
            to_remove.append(i)
    points = numpy.delete(points, to_remove, axis=1)
    lbounds = numpy.delete(lbounds, to_remove)
    ubounds = numpy.delete(ubounds, to_remove)

    if not len(lbounds):
        logger.warning("No dimension along which to compute the hypervolume.")
        return [0.] * len(pop)

    # Rescale the objective space
    # Note: 2 here is a magic number used to make the hypercube larger than it
    # really is. It makes sure that the individual always have a non-zero
    # hyper-volume contribution and improves the results while avoiding an
    # edge case.
    points = (points - lbounds) / numpy.max(ubounds.flatten())
    ubounds = numpy.max(points, axis=0) + 2.0

    hv = hype.hypeIndicatorSampled(
        points=points, bounds=ubounds, k=5, nrOfSamples=1000000
    )
    return hv


class CMA_MO(cma.StrategyMultiObjective):
    """Multiple objective covariance matrix adaption"""

    def __init__(
        self,
        centroids,
        offspring_size,
        sigma,
        max_ngen,
        IndCreator,
        RandIndCreator,
        weight_hv=0.5,
        map_function=None,
        use_scoop=False,
        use_stagnation_criterion=True,
    ):
        """Constructor

        Args:
            centroid (list): initial guess used as the starting point of
            the CMA-ES
            offspring_size (int): number of offspring individuals in each
                 generation
            sigma (float): initial standard deviation of the distribution
            max_ngen (int): total number of generation to run
            IndCreator (fcn): function returning an individual of the pop
            RandIndCreator (fcn): function creating a random individual.
            weight_hv (float): between 0 and 1. Weight given to the
                hypervolume contribution when computing the score of an
                individual in MO-CMA. The weight of the fitness contribution
                is computed as 1 - weight_hv.
            map_function (map): function used to map (parallelize) the
                 evaluation function calls
            use_scoop (bool): use scoop map for parallel computation
            use_stagnation_criterion (bool): whether to use the stagnation
                stopping criterion on top of the maximum generation criterion
        """

        if offspring_size is None:
            lambda_ = int(4 + 3 * log(len(RandIndCreator())))
        else:
            lambda_ = offspring_size

        if centroids is None:
            starters = [RandIndCreator() for i in range(lambda_)]
        else:
            if len(centroids) != lambda_:
                from itertools import cycle

                generator = cycle(centroids)
                starters = [
                    copy.deepcopy(next(generator)) for i in range(lambda_)
                ]
            else:
                starters = centroids

        cma.StrategyMultiObjective.__init__(
            self, starters, sigma, mu=int(lambda_ * 0.5), lambda_=lambda_
        )

        self.population = []
        self.problem_size = len(starters[0])

        self.weight_hv = weight_hv

        self.map_function = map_function
        self.use_scoop = use_scoop

        # Toolbox specific to this CMA-ES
        self.toolbox = base.Toolbox()
        self.toolbox.register("generate", self.generate, IndCreator)
        self.toolbox.register("update", self.update)

        if self.use_scoop:
            if self.map_function:
                raise Exception(
                    "Impossible to use scoop and provide self defined map "
                    "function: %s" % self.map_function
                )
            from scoop import futures

            self.map_function = futures.map

        # Set termination conditions
        self.active = True
        if max_ngen <= 0:
            max_ngen = 100 + 50 * (self.problem_size + 3) ** 2 / numpy.sqrt(
                lambda_
            )

        self.stopping_conditions = [
            MaxNGen(max_ngen),
        ]
        if use_stagnation_criterion:
            self.stopping_conditions.append(
                Stagnationv2(lambda_, self.problem_size)
            )

    def _select(self, candidates):
        """Select the best candidates of the population

         Fill the next population (chosen) with the Pareto fronts until there
         is not enough space. When an entire front does not fit in the space
         left we rely on a mixture of hypervolume and fitness. The respective
         weights of hypervolume and fitness are "hv" and "1-hv". The remaining
         fronts are explicitly not chosen"""

        if self.weight_hv == 0.0:
            fit = [numpy.sum(ind.fitness.values) for ind in candidates]
            idx_scores = list(numpy.argsort(fit))

        elif self.weight_hv == 1.0:
            hv = get_hyped(candidates)
            idx_scores = list(numpy.argsort(hv))[::-1]

        else:
            hv = get_hyped(candidates)
            idx_hv = list(numpy.argsort(hv))[::-1]
            fit = [numpy.sum(ind.fitness.values) for ind in candidates]
            idx_fit = list(numpy.argsort(fit))
            scores = []
            for i in range(len(candidates)):
                score = (self.weight_hv * idx_hv.index(i)) + (
                    (1.0 - self.weight_hv) * idx_fit.index(i)
                )
                scores.append(score)
            idx_scores = list(numpy.argsort(scores))

        chosen = [candidates[i] for i in idx_scores[: self.mu]]
        not_chosen = [candidates[i] for i in idx_scores[self.mu:]]
        return chosen, not_chosen

    def get_population(self, to_space):
        """Returns the population in the original parameter space"""
        pop = copy.deepcopy(self.population)
        for i, ind in enumerate(pop):
            for j, v in enumerate(ind):
                pop[i][j] = to_space[j](v)
        return pop

    def get_parents(self, to_space):
        """Returns the population in the original parameter space"""
        pop = copy.deepcopy(self.parents)
        for i, ind in enumerate(pop):
            for j, v in enumerate(ind):
                pop[i][j] = to_space[j](v)
        return pop

    def generate_new_pop(self, lbounds, ubounds):
        """Generate a new population bounded in the normalized space"""
        self.population = self.toolbox.generate()
        return utils.bound(self.population, lbounds, ubounds)

    def update_strategy(self):
        self.toolbox.update(self.population)

    def set_fitness(self, fitnesses):
        for f, ind in zip(fitnesses, self.population):
            ind.fitness.values = f

    def set_fitness_parents(self, fitnesses):
        for f, ind in zip(fitnesses, self.parents):
            ind.fitness.values = f

    def check_termination(self, gen):
        stopping_params = {
            "gen": gen,
            "population": self.population,
        }

        [c.check(stopping_params) for c in self.stopping_conditions]
        for c in self.stopping_conditions:
            if c.criteria_met:
                logger.info(
                    "CMA stopped because of termination criteria: " +
                    "" + " ".join(c.name)
                )
                self.active = False



"""bluepyopt.ephys.simulators tests"""

"""
Copyright (c) 2016-2020, EPFL/Blue Brain Project

 This file is part of BluePyOpt <https://github.com/BlueBrain/BluePyOpt>

 This library is free software; you can redistribute it and/or modify it under
 the terms of the GNU Lesser General Public License version 3.0 as published
 by the Free Software Foundation.

 This library is distributed in the hope that it will be useful, but WITHOUT
 ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
 FOR A PARTICULAR PURPOSE.  See the GNU Lesser General Public License for more
 details.

 You should have received a copy of the GNU Lesser General Public License
 along with this library; if not, write to the Free Software Foundation, Inc.,
 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.
"""

# pylint:disable=W0612


import pytest

import bluepyopt.ephys as ephys
from .testmodels import dummycells


@pytest.mark.unit
def test_distloc_exception():
    """ephys.protocols: test if protocol raise dist loc exception"""

    nrn_sim = ephys.simulators.NrnSimulator()
    dummy_cell = dummycells.DummyCellModel1()
    # icell = dummy_cell.instantiate(sim=nrn_sim)
    soma_loc = ephys.locations.NrnSeclistCompLocation(
        name='soma_loc',
        seclist_name='somatic',
        sec_index=0,
        comp_x=.5)
    dend_loc = ephys.locations.NrnSomaDistanceCompLocation(
        name='dend_loc',
        soma_distance=800,
        seclist_name='apical')

    rec_soma = ephys.recordings.CompRecording(
        name='soma.v',
        location=soma_loc,
        variable='v')
    rec_dend = ephys.recordings.CompRecording(
        name='dend.v',
        location=dend_loc,
        variable='v')

    stim = ephys.stimuli.NrnSquarePulse(
        step_amplitude=0.0,
        step_delay=0.0,
        step_duration=50,
        total_duration=50,
        location=soma_loc)

    protocol = ephys.protocols.SweepProtocol(
        name='prot',
        stimuli=[stim],
        recordings=[
            rec_soma,
            rec_dend])

    responses = protocol.run(
        cell_model=dummy_cell,
        param_values={},
        sim=nrn_sim)

    assert responses['soma.v'] is not None
    assert responses['dend.v'] is None

    protocol.destroy(sim=nrn_sim)
    dummy_cell.destroy(sim=nrn_sim)


def run_RuntimeError(
        self,
        tstop=None,
        dt=None,
        cvode_active=None,
        random123_globalindex=None):
    """Mock version of run that throws runtimeerror"""
    raise RuntimeError()


def run_NrnSimulatorException(
        self,
        tstop=None,
        dt=None,
        cvode_active=None,
        random123_globalindex=None):
    """Mock version of run that throws runtimeerror"""
    raise ephys.simulators.NrnSimulatorException('mock', None)


@pytest.mark.unit
def test_sweepprotocol_init():
    """ephys.protocols: Test SweepProtocol init"""

    nrn_sim = ephys.simulators.NrnSimulator()
    dummy_cell = dummycells.DummyCellModel1()
    # icell = dummy_cell.instantiate(sim=nrn_sim)
    soma_loc = ephys.locations.NrnSeclistCompLocation(
        name='soma_loc',
        seclist_name='somatic',
        sec_index=0,
        comp_x=.5)

    rec_soma = ephys.recordings.CompRecording(
        name='soma.v',
        location=soma_loc,
        variable='v')

    stim = ephys.stimuli.NrnSquarePulse(
        step_amplitude=0.0,
        step_delay=0.0,
        step_duration=50,
        total_duration=50,
        location=soma_loc)

    protocol = ephys.protocols.SweepProtocol(
        name='prot',
        stimuli=[stim],
        recordings=[rec_soma])

    assert isinstance(protocol, ephys.protocols.SweepProtocol)
    assert protocol.total_duration == 50
    assert (
        protocol.subprotocols() == {'prot': protocol})

    assert 'somatic[0](0.5)' in str(protocol)

    protocol.destroy(sim=nrn_sim)
    dummy_cell.destroy(sim=nrn_sim)


@pytest.mark.unit
def test_sequenceprotocol_init():
    """ephys.protocols: Test SequenceProtocol init"""

    nrn_sim = ephys.simulators.NrnSimulator()
    dummy_cell = dummycells.DummyCellModel1()
    # icell = dummy_cell.instantiate(sim=nrn_sim)
    soma_loc = ephys.locations.NrnSeclistCompLocation(
        name='soma_loc',
        seclist_name='somatic',
        sec_index=0,
        comp_x=.5)

    rec_soma = ephys.recordings.CompRecording(
        name='soma.v',
        location=soma_loc,
        variable='v')

    stim = ephys.stimuli.NrnSquarePulse(
        step_amplitude=0.0,
        step_delay=0.0,
        step_duration=50,
        total_duration=50,
        location=soma_loc)

    sweep_protocol = ephys.protocols.SweepProtocol(
        name='sweep_prot',
        stimuli=[stim],
        recordings=[rec_soma])

    seq_protocol = ephys.protocols.SequenceProtocol(
        name='seq_prot',
        protocols=[sweep_protocol])

    assert isinstance(seq_protocol, ephys.protocols.SequenceProtocol)
    assert (
        seq_protocol.subprotocols() == {
            'seq_prot': seq_protocol, 'sweep_prot': sweep_protocol})

    sweep_protocol.destroy(sim=nrn_sim)
    dummy_cell.destroy(sim=nrn_sim)


@pytest.mark.unit
def test_sequenceprotocol_run():
    """ephys.protocols: Test SequenceProtocol run"""

    nrn_sim = ephys.simulators.NrnSimulator()
    dummy_cell = dummycells.DummyCellModel1()
    # icell = dummy_cell.instantiate(sim=nrn_sim)
    soma_loc = ephys.locations.NrnSeclistCompLocation(
        name='soma_loc',
        seclist_name='somatic',
        sec_index=0,
        comp_x=.5)

    rec_soma = ephys.recordings.CompRecording(
        name='soma.v',
        location=soma_loc,
        variable='v')

    stim = ephys.stimuli.NrnSquarePulse(
        step_amplitude=0.0,
        step_delay=0.0,
        step_duration=50,
        total_duration=50,
        location=soma_loc)

    sweep_protocol = ephys.protocols.SweepProtocol(
        name='sweep_prot',
        stimuli=[stim],
        recordings=[rec_soma])

    seq_protocol = ephys.protocols.SequenceProtocol(
        name='seq_prot',
        protocols=[sweep_protocol])

    responses = seq_protocol.run(
        cell_model=dummy_cell,
        param_values={},
        sim=nrn_sim)

    assert responses is not None

    sweep_protocol.destroy(sim=nrn_sim)
    dummy_cell.destroy(sim=nrn_sim)


@pytest.mark.unit
def test_sequenceprotocol_overwrite():
    """ephys.protocols: Test SequenceProtocol overwriting keys"""

    nrn_sim = ephys.simulators.NrnSimulator()
    dummy_cell = dummycells.DummyCellModel1()

    sweep_protocols = []
    for x in [.2, .5]:
        soma_loc = ephys.locations.NrnSeclistCompLocation(
            name='soma_loc',
            seclist_name='somatic',
            sec_index=0,
            comp_x=x)

        rec_soma = ephys.recordings.CompRecording(
            name='soma.v',
            location=soma_loc,
            variable='v')

        stim = ephys.stimuli.NrnSquarePulse(
            step_amplitude=0.0,
            step_delay=0.0,
            step_duration=50,
            total_duration=50,
            location=soma_loc)

        sweep_protocols.append(ephys.protocols.SweepProtocol(
            name='sweep_prot',
            stimuli=[stim],
            recordings=[rec_soma]))

    seq_protocol = ephys.protocols.SequenceProtocol(
        name='seq_prot',
        protocols=sweep_protocols)

    pytest.raises(Exception, seq_protocol.run,
                  cell_model=dummy_cell,
                  param_values={},
                  sim=nrn_sim)

    for sweep_protocol in sweep_protocols:
        sweep_protocol.destroy(sim=nrn_sim)
    dummy_cell.destroy(sim=nrn_sim)


@pytest.mark.unit
def test_stepprotocol_init():
    """ephys.protocols: Test StepProtocol init"""

    soma_loc = ephys.locations.NrnSeclistCompLocation(
        name='soma_loc',
        seclist_name='somatic',
        sec_index=0,
        comp_x=.5)

    rec_soma = ephys.recordings.CompRecording(
        name='soma.v',
        location=soma_loc,
        variable='v')

    stim = ephys.stimuli.NrnSquarePulse(
        step_amplitude=0.0,
        step_delay=5.0,
        step_duration=50,
        total_duration=50,
        location=soma_loc)
    hold_stim = ephys.stimuli.NrnSquarePulse(
        step_amplitude=0.0,
        step_delay=0.0,
        step_duration=50,
        total_duration=50,
        location=soma_loc)

    step_protocol = ephys.protocols.StepProtocol(
        name='step_prot',
        step_stimulus=stim,
        holding_stimulus=hold_stim,
        recordings=[rec_soma])

    assert step_protocol.step_delay == 5.0
    assert step_protocol.step_duration == 50


@pytest.mark.unit
def test_sweepprotocol_run_unisolated():
    """ephys.protocols: Test SweepProtocol unisolated run"""

    nrn_sim = ephys.simulators.NrnSimulator()
    dummy_cell = dummycells.DummyCellModel1()
    # icell = dummy_cell.instantiate(sim=nrn_sim)
    soma_loc = ephys.locations.NrnSeclistCompLocation(
        name='soma_loc',
        seclist_name='somatic',
        sec_index=0,
        comp_x=.5)
    unknown_loc = ephys.locations.NrnSomaDistanceCompLocation(
        name='unknown_loc',
        seclist_name='somatic',
        soma_distance=100)

    rec_soma = ephys.recordings.CompRecording(
        name='soma.v',
        location=soma_loc,
        variable='v')
    rec_unknown = ephys.recordings.CompRecording(
        name='unknown.v',
        location=unknown_loc,
        variable='v')

    stim = ephys.stimuli.NrnSquarePulse(
        step_amplitude=0.0,
        step_delay=0.0,
        step_duration=50,
        total_duration=50,
        location=soma_loc)

    protocol = ephys.protocols.SweepProtocol(
        name='prot',
        stimuli=[stim],
        recordings=[rec_soma, rec_unknown])

    responses = protocol.run(
        cell_model=dummy_cell,
        param_values={},
        sim=nrn_sim,
        isolate=False)

    assert 'soma.v' in responses
    assert 'unknown.v' in responses
    assert responses['unknown.v'] is None

    protocol.destroy(sim=nrn_sim)
    dummy_cell.destroy(sim=nrn_sim)


@pytest.mark.unit
def test_sweepprotocol_run_isolated():
    """ephys.protocols: Test SweepProtocol isolated run"""

    nrn_sim = ephys.simulators.NrnSimulator(dt=0.1)
    dummy_cell = dummycells.DummyCellModel1()
    soma_loc = ephys.locations.NrnSeclistCompLocation(
        name='soma_loc',
        seclist_name='somatic',
        sec_index=0,
        comp_x=.5)
    unknown_loc = ephys.locations.NrnSomaDistanceCompLocation(
        name='unknown_loc',
        seclist_name='somatic',
        soma_distance=100)

    rec_soma = ephys.recordings.CompRecording(
        name='soma.v',
        location=soma_loc,
        variable='v')
    rec_unknown = ephys.recordings.CompRecording(
        name='unknown.v',
        location=unknown_loc,
        variable='v')

    stim = ephys.stimuli.NrnSquarePulse(
        step_amplitude=0.0,
        step_delay=0.0,
        step_duration=50,
        total_duration=50,
        location=soma_loc)

    protocol = ephys.protocols.SweepProtocol(
        name='prot',
        stimuli=[stim],
        recordings=[rec_soma, rec_unknown])

    responses = protocol.run(
        cell_model=dummy_cell,
        param_values={},
        sim=nrn_sim,
        isolate=True)

    assert 'soma.v' in responses
    assert 'unknown.v' in responses
    assert responses['unknown.v'] is None

    protocol.destroy(sim=nrn_sim)
    dummy_cell.destroy(sim=nrn_sim)


@pytest.mark.unit
def test_nrnsimulator_exception():
    """ephys.protocols: test if protocol raise nrn sim exception"""

    nrn_sim = ephys.simulators.NrnSimulator()
    dummy_cell = dummycells.DummyCellModel1()
    soma_loc = ephys.locations.NrnSeclistCompLocation(
        name='soma_loc',
        seclist_name='somatic',
        sec_index=0,
        comp_x=.5)

    rec_soma = ephys.recordings.CompRecording(
        name='soma.v',
        location=soma_loc,
        variable='v')

    stim = ephys.stimuli.NrnSquarePulse(
        step_amplitude=0.0,
        step_delay=0.0,
        step_duration=50,
        total_duration=50,
        location=soma_loc)

    protocol = ephys.protocols.SweepProtocol(
        name='prot',
        stimuli=[stim],
        recordings=[rec_soma])

    nrn_sim.run = run_RuntimeError

    responses = protocol.run(
        cell_model=dummy_cell,
        param_values={},
        sim=nrn_sim,
        isolate=False)

    assert responses['soma.v'] is None

    nrn_sim.run = run_NrnSimulatorException

    responses = protocol.run(
        cell_model=dummy_cell,
        param_values={},
        sim=nrn_sim,
        isolate=False)

    assert responses['soma.v'] is None

    protocol.destroy(sim=nrn_sim)
    dummy_cell.destroy(sim=nrn_sim)


@pytest.mark.unit
def test_sweepprotocol_instantiate_with_LFPyCellModel():
    """ephys.protocols: Test SweepProtocol instantiate with LFPyCellModel"""
    import LFPy

    dummy_cell = dummycells.DummyLFPyCellModel1()
    nrn_sim = ephys.simulators.LFPySimulator()

    soma_loc = ephys.locations.NrnSeclistCompLocation(
        name='soma_loc',
        seclist_name='somatic',
        sec_index=0,
        comp_x=.5
    )

    rec = ephys.recordings.LFPRecording(
        name='lfpy_rec'
    )

    stim = ephys.stimuli.LFPySquarePulse(
        step_amplitude=0.1,
        step_delay=70,
        step_duration=100,
        location=soma_loc,
        total_duration=300
    )

    protocol = ephys.protocols.SweepProtocol(
        name='prot',
        stimuli=[stim],
        recordings=[rec])

    dummy_cell.instantiate(sim=nrn_sim)
    protocol.instantiate(sim=nrn_sim, cell_model=dummy_cell)

    # check that recording and stimuli have been instantiated with LFPy
    assert rec.instantiated
    assert isinstance(rec.cell, LFPy.Cell)
    assert isinstance(stim.iclamp, LFPy.StimIntElectrode)

    protocol.destroy(sim=nrn_sim)
    dummy_cell.destroy(sim=nrn_sim)



"""Test neuroml functions"""

import os
import sys

import efel
import neuroml
import numpy
import pytest
from pyneuroml import pynml
from bluepyopt import ephys
from bluepyopt.neuroml import biophys
from bluepyopt.neuroml import cell
from bluepyopt.neuroml import morphology
from bluepyopt.neuroml import simulation

L5PC_PATH = os.path.abspath(
    os.path.join(os.path.dirname(__file__), "../../examples/l5pc")
)

sys.path.insert(0, L5PC_PATH)

import l5pc_evaluator
import l5pc_model  # NOQA

l5pc_cell = l5pc_model.create()
protocols = l5pc_evaluator.define_protocols()

release_params = {
    "gNaTs2_tbar_NaTs2_t.apical": 0.026145,
    "gSKv3_1bar_SKv3_1.apical": 0.004226,
    "gImbar_Im.apical": 0.000143,
    "gNaTa_tbar_NaTa_t.axonal": 3.137968,
    "gK_Tstbar_K_Tst.axonal": 0.089259,
    "gamma_CaDynamics_E2.axonal": 0.002910,
    "gNap_Et2bar_Nap_Et2.axonal": 0.006827,
    "gSK_E2bar_SK_E2.axonal": 0.007104,
    "gCa_HVAbar_Ca_HVA.axonal": 0.000990,
    "gK_Pstbar_K_Pst.axonal": 0.973538,
    "gSKv3_1bar_SKv3_1.axonal": 1.021945,
    "decay_CaDynamics_E2.axonal": 287.198731,
    "gCa_LVAstbar_Ca_LVAst.axonal": 0.008752,
    "gamma_CaDynamics_E2.somatic": 0.000609,
    "gSKv3_1bar_SKv3_1.somatic": 0.303472,
    "gSK_E2bar_SK_E2.somatic": 0.008407,
    "gCa_HVAbar_Ca_HVA.somatic": 0.000994,
    "gNaTs2_tbar_NaTs2_t.somatic": 0.983955,
    "decay_CaDynamics_E2.somatic": 210.485284,
    "gCa_LVAstbar_Ca_LVAst.somatic": 0.000333,
}


@pytest.mark.unit
def test_get_nml_mech_dir():
    """biophys.get_nml_mech_dir: Test get_nml_mech_dir"""
    channels = [
        "Ih",
        "NaTa_t",
        "NaTs2_t",
        "Nap_Et2",
        "K_Tst",
        "K_Pst",
        "SKv3_1",
        "SK_E2",
        "StochKv_deterministic",
        "KdShu2007",
        "Im",
        "Ca",
        "Ca_HVA",
        "Ca_LVAst",
        "pas",
    ]

    for channel in channels:
        assert os.path.isfile(
            os.path.join(biophys.get_nml_mech_dir(), f"{channel}.channel.nml")
        )
    assert os.path.isfile(
        os.path.join(biophys.get_nml_mech_dir(), "baseCaDynamics_E2_NML2.nml")
    )


@pytest.mark.unit
def test_get_channel_from_param_name():
    """biophys.get_channel_from_param_name:

    Test get_channel_from_param_name
    """
    # 3 underscores case
    assert biophys.get_channel_from_param_name(
        "gNaTs2_tbar_NaTs2_t"
    ) == "NaTs2_t"
    # 2 underscores case
    assert biophys.get_channel_from_param_name(
        "gamma_CaDynamics_E2"
    ) == "CaDynamics_E2"
    # 1 underscore case
    assert biophys.get_channel_from_param_name("gIhbar_Ih") == "Ih"


@pytest.mark.unit
def test_format_dist_fun():
    """biophys.format_dist_fun: Test format_dist_fun"""
    raw_expr = "(-0.8696 + 2.087*math.exp(({distance})*0.0031))*{value}"
    formatted_expr = "(-0.8696 + 2.087*exp(p*0.0031))*8e-05"
    assert biophys.format_dist_fun(raw_expr, 8e-05, None) == formatted_expr


@pytest.mark.unit
def test_add_nml_channel_to_nml_cell_file():
    """biophys.add_nml_channel_to_nml_cell_file:

    Test add_nml_channel_to_nml_cell_file
    """
    empty_cell_doc = neuroml.NeuroMLDocument(id="test_nml_cell")
    included_channels = []

    pytest.raises(
        ValueError,
        biophys.add_nml_channel_to_nml_cell_file,
        empty_cell_doc,
        included_channels,
    )

    channel_name = "K_Pst"
    biophys.add_nml_channel_to_nml_cell_file(
        empty_cell_doc,
        included_channels,
        channel_name=channel_name,
        skip_channels_copy=True,
    )
    assert len(empty_cell_doc.includes) == 1
    assert channel_name in str(empty_cell_doc.includes[0].href)
    assert len(included_channels) == 1
    assert f"{channel_name}.channel.nml" in included_channels

    # test that we cannot add the same channel twice
    biophys.add_nml_channel_to_nml_cell_file(
        empty_cell_doc,
        included_channels,
        channel_name=channel_name,
        skip_channels_copy=True,
    )
    assert len(empty_cell_doc.includes) == 1
    assert len(included_channels) == 1


@pytest.mark.unit
def test_get_arguments():
    """biophys.get_arguments: Test get_arguments"""
    parameter_name = "gSKv3_1bar_SKv3_1"
    section_list = "somatic"
    channel = "SKv3_1"
    variable_parameters = None
    cond_density = "0.303472 S_per_cm2"
    arguments, channel_class = biophys.get_arguments(
        l5pc_cell.params,
        parameter_name,
        section_list,
        channel,
        channel,
        variable_parameters,
        cond_density,
        release_params,
    )
    assert arguments["ion"] == "k"
    assert arguments["segment_groups"] == section_list
    assert arguments["erev"] == "-85.0 mV"
    assert arguments["id"] == "somatic_gSKv3_1bar_SKv3_1"
    assert arguments["ion_channel"] == channel
    assert arguments["cond_density"] == cond_density
    assert "variable_parameters" not in arguments
    assert channel_class == "ChannelDensity"

    # pas case
    parameter_name = "g_pas"
    section_list = "all"
    channel = "pas"
    cond_density = "3e-05 S_per_cm2"
    arguments, channel_class = biophys.get_arguments(
        l5pc_cell.params,
        parameter_name,
        section_list,
        channel,
        channel,
        variable_parameters,
        cond_density,
        release_params,
    )
    assert arguments["ion"] == "non_specific"
    assert arguments["segment_groups"] == section_list
    assert arguments["erev"] == "-75 mV"
    assert arguments["id"] == "all_g_pas"
    assert arguments["ion_channel"] == channel
    assert arguments["cond_density"] == cond_density
    assert "variable_parameters" not in arguments
    assert channel_class == "ChannelDensity"

    # nernst case
    parameter_name = "gCa_HVAbar_Ca_HVA"
    section_list = "axonal"
    channel = "Ca_HVA"
    cond_density = "0.00099 S_per_cm2"
    arguments, channel_class = biophys.get_arguments(
        l5pc_cell.params,
        parameter_name,
        section_list,
        channel,
        channel,
        variable_parameters,
        cond_density,
        release_params,
    )
    assert arguments["ion"] == "ca"
    assert arguments["segment_groups"] == section_list
    assert "erev" not in arguments
    assert arguments["id"] == "axonal_gCa_HVAbar_Ca_HVA"
    assert arguments["ion_channel"] == channel
    assert arguments["cond_density"] == cond_density
    assert "variable_parameters" not in arguments
    assert channel_class == "ChannelDensityNernst"


@pytest.mark.unit
def test_extract_parameter_value():
    """biophys.extract_parameter_value: Test extract_parameter_value"""
    # uniform parameter case
    param_name = "gSKv3_1bar_SKv3_1"
    section_list = "apical"
    channel = "SKv3_1"
    cond_density, variable_parameters = biophys.extract_parameter_value(
        l5pc_cell.params[".".join((param_name, section_list))],
        section_list,
        channel,
        True,
        release_params,
    )
    assert cond_density == "0.004226 S_per_cm2"
    assert variable_parameters is None

    # skipped non uniform parameter case
    param_name = "gIhbar_Ih"
    section_list = "apical"
    channel = "Ih"
    cond_density, variable_parameters = biophys.extract_parameter_value(
        l5pc_cell.params[".".join((param_name, section_list))],
        section_list,
        channel,
        True,
        release_params,
    )
    assert cond_density is None
    assert variable_parameters is None

    # non uniform parameter case (unskipped)
    param_name = "gIhbar_Ih"
    section_list = "apical"
    channel = "Ih"
    cond_density, variable_parameters = biophys.extract_parameter_value(
        l5pc_cell.params[".".join((param_name, section_list))],
        section_list,
        channel,
        False,
        release_params,
    )
    assert cond_density is None
    assert variable_parameters[0].segment_groups == section_list
    assert variable_parameters[0].parameter == "condDensity"
    assert (
        variable_parameters[0].inhomogeneous_value.inhomogeneous_parameters
        == "PathLengthOver_apical"
    )
    assert (
        variable_parameters[0].inhomogeneous_value.value
        == "(-0.8696 + 2.087*exp(p*0.0031))*8e-05"
    )


@pytest.mark.unit
def test_get_density():
    """biophys.get_density: Test get_density"""
    empty_cell_doc = neuroml.NeuroMLDocument(id="test_nml_cell")
    param_name = "gSK_E2bar_SK_E2"
    section_list = "axonal"
    density, channel_class = biophys.get_density(
        empty_cell_doc,
        l5pc_cell,
        l5pc_cell.params[".".join((param_name, section_list))],
        section_list,
        included_channels=[],
        skip_non_uniform=True,
        release_params=release_params,
        skip_channels_copy=True,
    )

    assert channel_class == "ChannelDensity"
    assert density.id == "axonal_gSK_E2bar_SK_E2"
    assert density.ion_channel == "SK_E2"
    assert density.cond_density == "0.007104 S_per_cm2"
    assert density.erev == "-85.0 mV"
    assert density.segment_groups == "axonal"
    assert density.ion == "k"


@pytest.mark.unit
def test_get_specific_capacitance():
    """biophys.get_specific_capacitance: Test get_specific_capacitance"""
    # case: default
    specific_capacitances = biophys.get_specific_capacitance({})
    assert neuroml.SpecificCapacitance(
        value="1.0 uF_per_cm2", segment_groups="axonal"
    ) in specific_capacitances
    assert neuroml.SpecificCapacitance(
        value="1.0 uF_per_cm2", segment_groups="somatic"
    ) in specific_capacitances
    assert neuroml.SpecificCapacitance(
        value="1.0 uF_per_cm2", segment_groups="basal"
    ) in specific_capacitances
    assert neuroml.SpecificCapacitance(
        value="1.0 uF_per_cm2", segment_groups="apical"
    ) in specific_capacitances

    # case: all
    specific_capacitances = biophys.get_specific_capacitance(
        {"all": "2.0 uF_per_cm2"}
    )
    assert neuroml.SpecificCapacitance(
        value="2.0 uF_per_cm2", segment_groups="axonal"
    ) in specific_capacitances
    assert neuroml.SpecificCapacitance(
        value="2.0 uF_per_cm2", segment_groups="somatic"
    ) in specific_capacitances
    assert neuroml.SpecificCapacitance(
        value="2.0 uF_per_cm2", segment_groups="basal"
    ) in specific_capacitances
    assert neuroml.SpecificCapacitance(
        value="2.0 uF_per_cm2", segment_groups="apical"
    ) in specific_capacitances

    # case: specific section(s)
    specific_capacitances = biophys.get_specific_capacitance(
        {"somatic": "2.0 uF_per_cm2", "axonal": "3.0 uF_per_cm2"}
    )
    assert neuroml.SpecificCapacitance(
        value="3.0 uF_per_cm2", segment_groups="axonal"
    ) in specific_capacitances
    assert neuroml.SpecificCapacitance(
        value="2.0 uF_per_cm2", segment_groups="somatic"
    ) in specific_capacitances
    assert neuroml.SpecificCapacitance(
        value="1.0 uF_per_cm2", segment_groups="basal"
    ) in specific_capacitances
    assert neuroml.SpecificCapacitance(
        value="1.0 uF_per_cm2", segment_groups="apical"
    ) in specific_capacitances


@pytest.mark.unit
def test_get_biophys():
    """biophys.get_biophys: Test get_biophys"""
    empty_cell_doc = neuroml.NeuroMLDocument(id="test_nml_cell")
    bio_prop = biophys.get_biophys(
        l5pc_cell,
        empty_cell_doc,
        release_params,
        skip_non_uniform=True,
        skip_channels_copy=True,
    )
    membrane_props = bio_prop.membrane_properties
    intracell_props = bio_prop.intracellular_properties

    assert bio_prop.id == "biophys"
    assert membrane_props.init_memb_potentials[0].value == "-65 mV"
    assert len(membrane_props.specific_capacitances) == 4
    assert len(membrane_props.channel_density_nernsts) == 4
    assert len(membrane_props.channel_densities) == 15
    assert intracell_props.resistivities[0].value == "100 ohm_cm"
    assert len(intracell_props.species) == 2


@pytest.mark.unit
def test_add_segment_groups():
    """morphology.add_segment_groups: Test add_segment_groups"""
    # creates a neuroml cell with a morphology
    cell = neuroml.Cell(id="nml_cell")
    morph = neuroml.Morphology(id="test_nml_cell_morph")
    for loc in ["soma", "axon", "dend", "apic"]:
        seg = neuroml.Segment(id=0, name=loc)
        morph.segments.append(seg)
    cell.morphology = morph

    morphology.add_segment_groups(cell)
    segment_group_names = [
        group.id for group in cell.morphology.segment_groups
    ]
    assert "somatic" in segment_group_names
    assert "axonal" in segment_group_names
    assert "basal" in segment_group_names
    assert "apical" in segment_group_names
    assert "soma_group" in segment_group_names
    assert "axon_group" in segment_group_names
    assert "dendrite_group" in segment_group_names


@pytest.mark.slow
@pytest.mark.neuroml
def test_neuroml_run():
    """Test neuroml conversion and simulation"""
    # replace axon is not supported yet
    l5pc_cell.morphology.do_replace_axon = False
    dt = 0.025
    protocol_name = "Step3"
    bpo_test_protocol = protocols[protocol_name]

    os.system("nrnivmodl examples/l5pc/mechanisms/")

    # create neuroml cell
    cell.create_neuroml_cell(
        l5pc_cell, release_params, skip_channels_copy=False
    )

    # create LEMS simulation
    network_filename = f"{l5pc_cell.name}.net.nml"
    lems_filename = f"LEMS_{l5pc_cell.name}.xml"
    simulation.create_neuroml_simulation(
        network_filename, bpo_test_protocol, dt, l5pc_cell.name, lems_filename
    )

    # remove compiled mechanisms if any before running LEMS simulation
    os.system("rm -rf x86_64/")

    # run the simulation with the NEURON simulator,
    # since the default jNeuroML simulator can only simulate single
    # compartment cells
    pynml.run_lems_with_jneuroml_neuron(
        lems_filename, nogui=True, plot=False)

    # re-compile the mechanisms before running with bluepyopt
    os.system("rm -rf x86_64/")
    os.system("nrnivmodl examples/l5pc/mechanisms/")

    lems_output = numpy.loadtxt("l5pc.Pop_l5pc_0_0.v.dat")
    lems_voltage = lems_output[:, 1] * 1000  # *1000 -> mV
    lems_time = lems_output[:, 0] * 1000  # *1000 -> ms

    # remove non uniform parameter to be coherent with LEMS simulation
    to_remove = []
    for key, param in l5pc_cell.params.items():
        if hasattr(param, "value_scaler"):
            if isinstance(
                param.value_scaler,
                ephys.parameterscalers.NrnSegmentSomaDistanceScaler,
            ):
                to_remove.append(key)

    for param_name in to_remove:
        l5pc_cell.params.pop(param_name)

    # run with regular bluepyopt
    sim = ephys.simulators.NrnSimulator(dt=dt, cvode_active=False)
    bpo_output = bpo_test_protocol.run(
        cell_model=l5pc_cell, param_values=release_params, sim=sim
    )

    response_name = f"{protocol_name}.soma.v"
    bpo_voltage = bpo_output[response_name]["voltage"]
    bpo_time = bpo_output[response_name]["time"]

    # we don't expect the two traces to be exactly the same,
    # but to be pretty similar
    # we consider it satisfactory if they fire within 3 ms respectively
    trace_lems = {
        "T": lems_time,
        "V": lems_voltage,
        "stim_start": [700],
        "stim_end": [2700]
    }
    trace_bpo = {
        "T": bpo_time,
        "V": bpo_voltage,
        "stim_start": [700],
        "stim_end": [2700]
    }
    traces = [trace_lems, trace_bpo]
    features = ["peak_time"]
    feature_values = efel.getFeatureValues(
        traces, features, raise_warnings=False
    )
    lems_peak_time = feature_values[0]["peak_time"]
    bpo_peak_time = feature_values[1]["peak_time"]

    numpy.testing.assert_allclose(lems_peak_time, bpo_peak_time, atol=3)



"""bluepyopt.ephys.simulators tests"""

"""
Copyright (c) 2016-2020, EPFL/Blue Brain Project

 This file is part of BluePyOpt <https://github.com/BlueBrain/BluePyOpt>

 This library is free software; you can redistribute it and/or modify it under
 the terms of the GNU Lesser General Public License version 3.0 as published
 by the Free Software Foundation.

 This library is distributed in the hope that it will be useful, but WITHOUT
 ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
 FOR A PARTICULAR PURPOSE.  See the GNU Lesser General Public License for more
 details.

 You should have received a copy of the GNU Lesser General Public License
 along with this library; if not, write to the Free Software Foundation, Inc.,
 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.
"""

# pylint:disable=W0612


import os

import pytest

import bluepyopt.ephys as ephys


@pytest.mark.unit
def test_comprecording_init():
    """ephys.recordings: Test CompRecording init"""

    recording = ephys.recordings.CompRecording()

    assert isinstance(recording, ephys.recordings.CompRecording)

    assert recording.response is None

    '''
    nrn_sim = ephys.simulators.NrnSimulator()
    dummy_cell = dummycells.DummyCellModel1()
    # icell = dummy_cell.instantiate(sim=nrn_sim)
    soma_loc = ephys.locations.NrnSeclistCompLocation(
        name='soma_loc',
        seclist_name='somatic',
        sec_index=0,
        comp_x=.5)

    rec_soma = ephys.recordings.CompRecording(
        name='soma.v',
        location=soma_loc,
        variable='v')

    stim = ephys.stimuli.NrnSquarePulse(
        step_amplitude=0.0,
        step_delay=0.0,
        step_duration=50,
        total_duration=50,
        location=soma_loc)

    protocol = ephys.protocols.SweepProtocol(
        name='prot',
        stimuli=[stim],
        recordings=[rec_soma])

    assert_true(isinstance(protocol, ephys.protocols.SweepProtocol))
    assert_equal(protocol.total_duration, 50)
    assert_equal(
        protocol.subprotocols(), {'prot': protocol})

    assert_true('somatic[0](0.5)' in str(protocol))

    protocol.destroy(sim=nrn_sim)
    dummy_cell.destroy(sim=nrn_sim)
    '''


@pytest.mark.unit
def test_lfprecording_init():
    """ephys.recordings: Test LFPRecording init"""

    recording = ephys.recordings.LFPRecording(name="rec")

    assert isinstance(recording, ephys.recordings.LFPRecording)

    assert recording.response is None
    assert str(recording) == "rec: v at extracellular"


@pytest.mark.unit
def test_lfprecording_instantiate():
    """ephys.recordings: Test LFPRecording instantiate"""
    TESTDATA_DIR = os.path.join(
        os.path.dirname(os.path.abspath(__file__)), 'testdata'
    )
    simple_morphology_path = os.path.join(TESTDATA_DIR, 'simple.swc')
    test_morph = ephys.morphologies.NrnFileMorphology(simple_morphology_path)

    recording = ephys.recordings.LFPRecording()
    lfpy_cell = ephys.models.LFPyCellModel(
        name="lfpy_cell", morph=test_morph, mechs=[]
    )
    neuron_sim = ephys.simulators.LFPySimulator()
    lfpy_cell.instantiate(sim=neuron_sim)

    recording.instantiate(
        sim=neuron_sim, lfpy_cell=lfpy_cell.lfpy_cell
    )

    assert recording.instantiated

    lfpy_cell.destroy(sim=neuron_sim)
